########
#
# postprocess_batch_results.py
#
# Given a .json or .csv file representing the output from the batch detection API,
# do one or more of the following:
# 
# * Evaluate detector precision/recall, optionally rendering results (requires
#   ground truth)
# * Sample true/false positives/negatives and render to HTML (requires ground
#   truth)
# * Sample detections/non-detections and render to HTML (when ground truth isn't
#   available)
# 
# Ground truth, if available, must be in the COCO Camera Traps format.
#
########

#%% Constants and imports

import argparse
import collections
import copy
import errno
import io
import itertools
import os
import sys
import time
import uuid
import urllib
import warnings

from typing import Any, Dict, Iterable, Optional, Tuple
from enum import IntEnum
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
import humanfriendly
import pandas as pd

from sklearn.metrics import precision_recall_curve, confusion_matrix, average_precision_score
from tqdm import tqdm

import md_visualization.visualization_utils as vis_utils
import md_visualization.plot_utils as plot_utils

from md_utils.write_html_image_list import write_html_image_list
from md_utils import path_utils
from data_management.cct_json_utils import (CameraTrapJsonUtils, IndexedJsonDb)
from api.batch_processing.postprocessing.load_api_results import load_api_results
from md_utils.ct_utils import args_to_object

from detection.run_detector import get_typical_confidence_threshold_from_results

warnings.filterwarnings('ignore', '(Possibly )?corrupt EXIF data', UserWarning)


#%% Options

DEFAULT_NEGATIVE_CLASSES = ['empty']
DEFAULT_UNKNOWN_CLASSES = ['unknown', 'unlabeled', 'ambiguous']


def has_overlap(set1: Iterable, set2: Iterable) -> bool:
    """
    Check whether two sets overlap.
    """
    
    return not set(set1).isdisjoint(set(set2))


# Make sure there is no overlap between the two sets, because this will cause
# issues in the code
assert not has_overlap(DEFAULT_NEGATIVE_CLASSES, DEFAULT_UNKNOWN_CLASSES), (
        'Default negative and unknown classes cannot overlap.')


class PostProcessingOptions:

    ### Required inputs

    api_output_file = ''
    output_dir = ''

    ### Options

    # Can be a folder or a SAS URL
    image_base_dir = '.'

    ground_truth_json_file = ''

    ## These apply only when we're doing ground-truth comparisons
    
    # Classes we'll treat as negative
    #
    # Include the token "#NO_LABELS#" to indicate that an image with no annotations
    # should be considered empty.
    negative_classes = DEFAULT_NEGATIVE_CLASSES
    
    # Classes we'll treat as neither positive nor negative
    unlabeled_classes = DEFAULT_UNKNOWN_CLASSES

    # A list of output sets that we should count, but not render images for.
    #
    # Typically used to preview sets with lots of empties, where you don't want to
    # subset but also don't want to render 100,000 empty images.
    #
    # detections, non_detections
    # detections_animal, detections_person, detections_vehicle
    rendering_bypass_sets = []

    # By default, choose a confidence threshold based on the detector version
    confidence_threshold = None
    classification_confidence_threshold = 0.5

    # Used for summary statistics only
    target_recall = 0.9

    # Number of images to sample, -1 for "all images"
    num_images_to_sample = 500

    # Random seed for sampling, or None
    sample_seed: Optional[int] = 0 # None

    viz_target_width = 800

    line_thickness = 4
    box_expansion = 0

    job_name_string = None
    model_version_string = None
    
    # These should really be mutually exclusive, but I'm not enforcing this.
    #
    # Nothing bad happens if you set both to true; the confidence sort happens
    # second.
    sort_html_by_filename = True
    sort_html_by_confidence = False

    link_images_to_originals = True
    
    # Optionally separate detections into categories (animal/vehicle/human)
    # 
    # Currently only supported when ground truth is unavailable
    separate_detections_by_category = True

    # Optionally replace one or more strings in filenames with other strings;
    # useful for taking a set of results generated for one folder structure
    # and applying them to a slightly different folder structure.
    api_output_filename_replacements = {}
    ground_truth_filename_replacements = {}

    # Allow bypassing API output loading when operating on previously-loaded
    # results
    api_detection_results: Optional[pd.DataFrame] = None
    api_other_fields: Optional[Dict[str, Any]] = None

    # Should we also split out a separate report about the detections that were
    # just below our main confidence threshold?
    #
    # Currently only supported when ground truth is unavailable
    include_almost_detections = False
    almost_detection_confidence_threshold = None

    # Control rendering parallelization
    parallelize_rendering_n_cores: Optional[int] = 100
    parallelize_rendering_with_threads = True
    parallelize_rendering = False
    
    sort_classification_results_by_count = False    
    
    # Should we split individual pages up into smaller pages if there are more than
    # N images?
    max_figures_per_html_file = None
    
# ...PostProcessingOptions


class PostProcessingResults:

    output_html_file = ''
    api_detection_results: Optional[pd.DataFrame] = None
    api_other_fields: Optional[Dict[str, Any]] = None


##%% Helper classes and functions

class DetectionStatus(IntEnum):
    """
    Flags used to mark images as positive or negative for P/R analysis
    (according to ground truth and/or detector output)
    """
    
    DS_NEGATIVE = 0
    DS_POSITIVE = 1

    # Anything greater than this isn't clearly positive or negative
    DS_MAX_DEFINITIVE_VALUE = DS_POSITIVE

    # image has annotations suggesting both negative and positive
    DS_AMBIGUOUS = 2

    # image is not annotated or is annotated with 'unknown', 'unlabeled', ETC.
    DS_UNKNOWN = 3

    # image has not yet been assigned a state
    DS_UNASSIGNED = 4

    # In some analyses, we add an additional class that lets us look at
    # detections just below our main confidence threshold
    DS_ALMOST = 5


def mark_detection_status(
        indexed_db: IndexedJsonDb,
        negative_classes: Iterable[str] = DEFAULT_NEGATIVE_CLASSES,
        unknown_classes: Iterable[str] = DEFAULT_UNKNOWN_CLASSES
        ) -> Tuple[int, int, int, int]:
    """
    For each image in indexed_db.db['images'], add a '_detection_status' field
    to indicate whether to treat this image as positive, negative, ambiguous,
    or unknown.

    Makes modifications in-place.

    returns (n_negative, n_positive, n_unknown, n_ambiguous)
    """
    
    negative_classes = set(negative_classes)
    unknown_classes = set(unknown_classes)

    # count the # of images with each type of DetectionStatus
    n_unknown = 0
    n_ambiguous = 0
    n_positive = 0
    n_negative = 0

    print('Preparing ground-truth annotations')
    for im in tqdm(indexed_db.db['images']):

        image_id = im['id']
        annotations = indexed_db.image_id_to_annotations[image_id]
        categories = [ann['category_id'] for ann in annotations]
        category_names = set(indexed_db.cat_id_to_name[cat] for cat in categories)

        # Check whether this image has:
        # - unknown / unassigned-type labels
        # - negative-type labels
        # - positive labels (i.e., labels that are neither unknown nor negative)
        has_unknown_labels = has_overlap(category_names, unknown_classes)
        has_negative_labels = has_overlap(category_names, negative_classes)
        has_positive_labels = 0 < len(category_names - (unknown_classes | negative_classes))
        # assert has_unknown_labels is False, '{} has unknown labels'.format(annotations)

        # If there are no image annotations...
        if len(categories) == 0:
                        
            if '#NO_LABELS#' in negative_classes:
                n_negative += 1
                im['_detection_status'] = DetectionStatus.DS_NEGATIVE
            else:
                n_unknown += 1
                im['_detection_status'] = DetectionStatus.DS_UNKNOWN

            # n_negative += 1
            # im['_detection_status'] = DetectionStatus.DS_NEGATIVE

        # If the image has more than one type of labels, it's ambiguous
        # note: bools are automatically converted to 0/1, so we can sum
        elif (has_unknown_labels + has_negative_labels + has_positive_labels) > 1:
            n_ambiguous += 1
            im['_detection_status'] = DetectionStatus.DS_AMBIGUOUS

        # After the check above, we can be sure it's only one of positive,
        # negative, or unknown.
        #
        # Important: do not merge the following 'unknown' branch with the first
        # 'unknown' branch above, where we tested 'if len(categories) == 0'
        #
        # If the image has only unknown labels
        elif has_unknown_labels:
            n_unknown += 1
            im['_detection_status'] = DetectionStatus.DS_UNKNOWN

        # If the image has only negative labels
        elif has_negative_labels:
            n_negative += 1
            im['_detection_status'] = DetectionStatus.DS_NEGATIVE

        # If the images has only positive labels
        elif has_positive_labels:
            n_positive += 1
            im['_detection_status'] = DetectionStatus.DS_POSITIVE

            # Annotate the category, if it is unambiguous
            if len(category_names) == 1:
                im['_unambiguous_category'] = list(category_names)[0]

        else:
            raise Exception('Invalid detection state')

    # ...for each image

    return n_negative, n_positive, n_unknown, n_ambiguous

# ...mark_detection_status()


def is_sas_url(s: str) -> bool:
    """
    Placeholder for a more robust way to verify that a link is a SAS URL.
    99.999% of the time this will suffice for what we're using it for right now.
    """
    
    return (s.startswith(('http://', 'https://')) and ('core.windows.net' in s)
            and ('?' in s))


def relative_sas_url(folder_url: str, relative_path: str) -> Optional[str]:
    """
    Given a container-level or folder-level SAS URL, create a SAS URL to the
    specified relative path.
    """
    
    relative_path = relative_path.replace('%','%25')
    relative_path = relative_path.replace('#','%23')
    relative_path = relative_path.replace(' ','%20')

    if not is_sas_url(folder_url):
        return None
    tokens = folder_url.split('?')
    assert len(tokens) == 2
    if not tokens[0].endswith('/'):
        tokens[0] = tokens[0] + '/'
    if relative_path.startswith('/'):
        relative_path = relative_path[1:]
    return tokens[0] + relative_path + '?' + tokens[1]


def render_bounding_boxes(
        image_base_dir,
        image_relative_path,
        display_name,
        detections,
        res,
        ground_truth_boxes=None,
        detection_categories=None,
        classification_categories=None,
        options=None):
    """
    Renders detection bounding boxes on a single image.

    The source image is:

        image_base_dir / image_relative_path

    The target image is, for example:

        [options.output_dir] / 
        ['detections' or 'non_detections'] / 
        [filename with slashes turned into tildes]

    "res" is a result type, e.g. "detections", "non-detections"; this determines the
    output folder for the rendered image.
    
    Only very preliminary support is provided for ground truth box rendering.
    
    Returns the html info struct for this image in the format that's used for
    write_html_image_list.
    """

    if options is None:
        options = PostProcessingOptions()

    # Leaving code in place for reading from blob storage, may support this
    # in the future.
    """
    stream = io.BytesIO()
    _ = blob_service.get_blob_to_stream(container_name, image_id, stream)
    # resize is to display them in this notebook or in the HTML more quickly
    image = Image.open(stream).resize(viz_size)
    """

    image_full_path = None
    
    if res in options.rendering_bypass_sets:

        sample_name = res + '_' + path_utils.flatten_path(image_relative_path)

    else:

        if is_sas_url(image_base_dir):
            image_full_path = relative_sas_url(image_base_dir, image_relative_path)
        else:
            image_full_path = os.path.join(image_base_dir, image_relative_path)

        # os.path.isfile() is slow when mounting remote directories; much faster
        # to just try/except on the image open.
        try:
            image = vis_utils.open_image(image_full_path)
        except:
            print('Warning: could not open image file {}'.format(image_full_path))
            image = None
            # return ''
        
        # Render images to a flat folder... we can use os.sep here because we've
        # already normalized paths
        sample_name = res + '_' + path_utils.flatten_path(image_relative_path)
        fullpath = os.path.join(options.output_dir, res, sample_name)

        if image is not None:
            
            original_size = image.size
            
            if options.viz_target_width is not None:
                image = vis_utils.resize_image(image, options.viz_target_width)
    
            if ground_truth_boxes is not None and len(ground_truth_boxes) > 0:
                
                # Create class labels like "gt_1" or "gt_27"
                gt_classes = [0] * len(ground_truth_boxes)
                label_map = {0:'ground truth'}
                # for i_box,box in enumerate(ground_truth_boxes):
                #    gt_classes.append('_' + str(box[-1]))
                vis_utils.render_db_bounding_boxes(ground_truth_boxes, gt_classes, image,
                                                   original_size=original_size,label_map=label_map,
                                                   thickness=4,expansion=4)
                
            vis_utils.render_detection_bounding_boxes(
                detections, image,
                label_map=detection_categories,
                classification_label_map=classification_categories,
                confidence_threshold=options.confidence_threshold,
                thickness=options.line_thickness,
                expansion=options.box_expansion)
    
            try:
                image.save(fullpath)
            except OSError as e:
                # errno.ENAMETOOLONG doesn't get thrown properly on Windows, so
                # we awkwardly check against a hard-coded limit
                if (e.errno == errno.ENAMETOOLONG) or (len(fullpath) >= 259):
                    extension = os.path.splitext(sample_name)[1]
                    sample_name = res + '_' + str(uuid.uuid4()) + extension
                    image.save(os.path.join(options.output_dir, res, sample_name))
                else:
                    raise

    # Use slashes regardless of os
    file_name = '{}/{}'.format(res,sample_name)

    info = {
        'filename': file_name,
        'title': display_name,
        'textStyle':\
         'font-family:verdana,arial,calibri;font-size:80%;text-align:left;margin-top:20;margin-bottom:5'
    }
    
    # Optionally add links back to the original images
    if options.link_images_to_originals and (image_full_path is not None):
        info['linkTarget'] = urllib.parse.quote(image_full_path)
        
    return info

# ...render_bounding_boxes


def prepare_html_subpages(images_html, output_dir, options=None):
    """
    Write out a series of html image lists, e.g. the fp/tp/fn/tn pages.

    image_html is a dictionary mapping an html page name (e.g. "fp") to a list
    of image structs friendly to write_html_image_list
    """
    
    if options is None:
        options = PostProcessingOptions()

    # Count items in each category
    image_counts = {}
    for res, array in images_html.items():
        image_counts[res] = len(array)

    # Optionally sort by filename before writing to html
    if options.sort_html_by_filename:
        images_html_sorted = {}
        for res, array in images_html.items():
            sorted_array = sorted(array, key=lambda x: x['filename'])
            images_html_sorted[res] = sorted_array
        images_html = images_html_sorted
    
    # Optionally sort by confidence before writing to html
    if options.sort_html_by_confidence:
        images_html_sorted = {}
        for res, array in images_html.items():
            
            if not all(['max_conf' in d for d in array]):
                print("Warning: some elements in the {} page don't have confidence values, can't sort by confidence".format(
                    res))
            else:
                sorted_array = sorted(array, key=lambda x: x['max_conf'], reverse=True)
                images_html_sorted[res] = sorted_array
                images_html = images_html_sorted
        
    # Write the individual HTML files
    for res, array in images_html.items():
        
        html_image_list_options = {}    
        html_image_list_options['maxFiguresPerHtmlFile'] = options.max_figures_per_html_file
        html_image_list_options['headerHtml'] = '<h1>{}</h1>'.format(res.upper())
        
        write_html_image_list(
            filename=os.path.join(output_dir, '{}.html'.format(res)),
            images=array,
            options=html_image_list_options)

    return image_counts

# ...prepare_html_subpages()

# Get unique categories above the threshold for this image
def get_positive_categories(detections,options):
    positive_categories = set()
    for d in detections:
        if d['conf'] >= options.confidence_threshold:
            positive_categories.add(d['category'])
    return sorted(positive_categories)


# Render an image (with no ground truth information)
#
# Returns a list of rendering structs, where the first item is a category (e.g. "detections_animal"), 
# and  the second is a dict of information needed for rendering.  E.g.:
#
# [['detections_animal', 
# {
#  'filename': 'detections_animal/detections_animal_blah~01060415.JPG', 
#  'title': '<b>Result type</b>: detections_animal, 
#            <b>Image</b>: blah\\01060415.JPG,
#            <b>Max conf</b>: 0.897',
#   'textStyle': 'font-family:verdana,arial,calibri;font-size:80%;text-align:left;margin-top:20;margin-bottom:5',
#   'linkTarget': 'full_path_to_%5C01060415.JPG'
# }]]
# 
# When no classification data is present, this list will always be length-1.  When
# classification data is present, an image may appear in multiple categories.
#
# Populates the 'max_conf' field of the first element of the list.
#
# Returns None if there are any errors.
def render_image_no_gt(file_info,detection_categories_to_results_name,
                       detection_categories,classification_categories,
                       options):

    image_relative_path = file_info[0]
    max_conf = file_info[1]
    detections = file_info[2]

    detection_status = DetectionStatus.DS_UNASSIGNED
    if max_conf >= options.confidence_threshold:
        detection_status = DetectionStatus.DS_POSITIVE
    else:
        if options.include_almost_detections:
            if max_conf >= options.almost_detection_confidence_threshold:
                detection_status = DetectionStatus.DS_ALMOST
            else:
                detection_status = DetectionStatus.DS_NEGATIVE
        else:
            detection_status = DetectionStatus.DS_NEGATIVE

    if detection_status == DetectionStatus.DS_POSITIVE:
        if options.separate_detections_by_category:
            positive_categories = tuple(get_positive_categories(detections,options))            
            if positive_categories not in detection_categories_to_results_name:
                raise ValueError('Error: {} not in category mapping (file {})'.format(
                    str(positive_categories),image_relative_path))
            res = detection_categories_to_results_name[positive_categories]
        else:
            res = 'detections'

    elif detection_status == DetectionStatus.DS_NEGATIVE:
        res = 'non_detections'
    else:
        assert detection_status == DetectionStatus.DS_ALMOST
        res = 'almost_detections'

    display_name = '<b>Result type</b>: {}, <b>Image</b>: {}, <b>Max conf</b>: {:0.3f}'.format(
        res, image_relative_path, max_conf)

    rendering_options = copy.copy(options)
    if detection_status == DetectionStatus.DS_ALMOST:
        rendering_options.confidence_threshold = \
            rendering_options.almost_detection_confidence_threshold
            
    rendered_image_html_info = render_bounding_boxes(
        image_base_dir=options.image_base_dir,
        image_relative_path=image_relative_path,
        display_name=display_name,
        detections=detections,
        res=res,
        ground_truth_boxes=None,
        detection_categories=detection_categories,
        classification_categories=classification_categories,
        options=rendering_options)

    image_result = None

    if len(rendered_image_html_info) > 0:

        image_result = [[res, rendered_image_html_info]]

        max_conf = 0
        
        for det in detections:
            
            if det['conf'] > max_conf:
                max_conf = det['conf']

            if ('classifications' in det):

                # This is a list of [class,confidence] pairs, sorted by confidence
                classifications = det['classifications']
                top1_class_id = classifications[0][0]
                top1_class_name = classification_categories[top1_class_id]
                top1_class_score = classifications[0][1]

                # If we either don't have a confidence threshold, or we've met our
                # confidence threshold
                if (options.classification_confidence_threshold < 0) or \
                    (top1_class_score >= options.classification_confidence_threshold):
                    image_result.append(['class_{}'.format(top1_class_name),
                                         rendered_image_html_info])
                else:
                    image_result.append(['class_unreliable',
                                         rendered_image_html_info])

            # ...if this detection has classification info

        # ...for each detection

        image_result[0][1]['max_conf'] = max_conf
        
    # ...if we got valid rendering info back from render_bounding_boxes()
    
    return image_result

# ...def render_image_no_gt()
    

# Render an image with ground truth information.  See render_image_no_gt for return
# data format.
def render_image_with_gt(file_info,ground_truth_indexed_db,
                         detection_categories,classification_categories,options):

    image_relative_path = file_info[0]
    max_conf = file_info[1]
    detections = file_info[2]

    # This should already have been normalized to either '/' or '\'

    image_id = ground_truth_indexed_db.filename_to_id.get(image_relative_path, None)
    if image_id is None:
        print('Warning: couldn''t find ground truth for image {}'.format(image_relative_path))
        return None

    image = ground_truth_indexed_db.image_id_to_image[image_id]
    annotations = ground_truth_indexed_db.image_id_to_annotations[image_id]

    ground_truth_boxes = []
    for ann in annotations:
        if 'bbox' in ann:
            ground_truth_box = [x for x in ann['bbox']]
            ground_truth_box.append(ann['category_id'])
            ground_truth_boxes.append(ground_truth_box)
    
    gt_status = image['_detection_status']

    gt_presence = bool(gt_status)

    gt_classes = CameraTrapJsonUtils.annotations_to_classnames(
        annotations, ground_truth_indexed_db.cat_id_to_name)
    gt_class_summary = ','.join(gt_classes)

    if gt_status > DetectionStatus.DS_MAX_DEFINITIVE_VALUE:
        print(f'Skipping image {image_id}, does not have a definitive '
              f'ground truth status (status: {gt_status}, classes: {gt_class_summary})')
        return None

    detected = max_conf > options.confidence_threshold

    if gt_presence and detected:
        if '_classification_accuracy' not in image.keys():
            res = 'tp'
        elif np.isclose(1, image['_classification_accuracy']):
            res = 'tpc'
        else:
            res = 'tpi'
    elif not gt_presence and detected:
        res = 'fp'
    elif gt_presence and not detected:
        res = 'fn'
    else:
        res = 'tn'

    display_name = '<b>Result type</b>: {}, <b>Presence</b>: {}, <b>Class</b>: {}, <b>Max conf</b>: {:0.3f}%, <b>Image</b>: {}'.format(
        res.upper(), str(gt_presence), gt_class_summary,
        max_conf * 100, image_relative_path)

    rendered_image_html_info = render_bounding_boxes(
        image_base_dir=options.image_base_dir,
        image_relative_path=image_relative_path,
        display_name=display_name,
        detections=detections,
        res=res,
        ground_truth_boxes=ground_truth_boxes,
        detection_categories=detection_categories,
        classification_categories=classification_categories,
        options=options)

    image_result = None
    if len(rendered_image_html_info) > 0:
        image_result = [[res, rendered_image_html_info]]
        for gt_class in gt_classes:
            image_result.append(['class_{}'.format(gt_class), rendered_image_html_info])

    return image_result

# ...def render_image_with_gt()

    
#%% Main function

def process_batch_results(options: PostProcessingOptions
                          ) -> PostProcessingResults:

    ppresults = PostProcessingResults()

    ##%% Expand some options for convenience

    output_dir = options.output_dir


    ##%% Prepare output dir

    os.makedirs(output_dir, exist_ok=True)


    ##%% Load ground truth if available

    ground_truth_indexed_db = None

    if (options.ground_truth_json_file is not None) and (len(options.ground_truth_json_file) > 0):

        if options.separate_detections_by_category:
            print("Warning: I don't know how to separate categories yet when doing " + \
                  "a P/R analysis, disabling category separation")
            options.separate_detections_by_category = False

        ground_truth_indexed_db = IndexedJsonDb(
            options.ground_truth_json_file, b_normalize_paths=True,
            filename_replacements=options.ground_truth_filename_replacements)

        # Mark images in the ground truth as positive or negative
        n_negative, n_positive, n_unknown, n_ambiguous = mark_detection_status(
            ground_truth_indexed_db, negative_classes=options.negative_classes,
            unknown_classes=options.unlabeled_classes)
        print(f'Finished loading and indexing ground truth: {n_negative} '
              f'negative, {n_positive} positive, {n_unknown} unknown, '
              f'{n_ambiguous} ambiguous')


    ##%% Load detection (and possibly classification) results

    # If the caller hasn't supplied results, load them
    if options.api_detection_results is None:
        detections_df, other_fields = load_api_results(
            options.api_output_file, normalize_paths=True,
            filename_replacements=options.api_output_filename_replacements)
        ppresults.api_detection_results = detections_df
        ppresults.api_other_fields = other_fields        

    else:
        print('Bypassing detection results loading...')
        assert options.api_other_fields is not None
        detections_df = options.api_detection_results
        other_fields = options.api_other_fields

    # Determine confidence thresholds if necessary
    
    if options.confidence_threshold is None:
        options.confidence_threshold = \
            get_typical_confidence_threshold_from_results(other_fields)
        print('Choosing default confidence threshold of {} based on MD version'.format(
            options.confidence_threshold))    
            
    if options.almost_detection_confidence_threshold is None:
        options.almost_detection_confidence_threshold = options.confidence_threshold - 0.05
        if options.almost_detection_confidence_threshold < 0:
            options.almost_detection_confidence_threshold = 0
            
    # Remove failed rows
    n_failures = 0
    if 'failure' in detections_df.columns:
        n_failures = detections_df['failure'].count()
        print('Ignoring {} failed images'.format(n_failures))
        # Explicitly forcing a copy() operation here to suppress "trying to be set
        # on a copy" # warnings (and associated risks) below.
        detections_df = detections_df[detections_df['failure'].isna()].copy()
        
    assert other_fields is not None

    detection_categories = other_fields['detection_categories']
    
    # Convert keys and values to lowercase
    classification_categories = other_fields.get('classification_categories', {})
    if classification_categories is not None:
        classification_categories = {
            k.lower(): v.lower()
            for k, v in classification_categories.items()
        }

    # Add column 'pred_detection_label' to indicate predicted detection status,
    # not separating out the classes
    det_status = 'pred_detection_label'
    if options.include_almost_detections:
        detections_df[det_status] = DetectionStatus.DS_ALMOST
        confidences = detections_df['max_detection_conf']

        pos_mask = (confidences >= options.confidence_threshold)
        detections_df.loc[pos_mask, det_status] = DetectionStatus.DS_POSITIVE

        neg_mask = (confidences < options.almost_detection_confidence_threshold)
        detections_df.loc[neg_mask, det_status] = DetectionStatus.DS_NEGATIVE
    else:
        detections_df[det_status] = np.where(
            detections_df['max_detection_conf'] >= options.confidence_threshold,
            DetectionStatus.DS_POSITIVE, DetectionStatus.DS_NEGATIVE)

    n_positives = sum(detections_df[det_status] == DetectionStatus.DS_POSITIVE)
    print(f'Finished loading and preprocessing {len(detections_df)} rows '
          f'from detector output, predicted {n_positives} positives.')

    if options.include_almost_detections:
        n_almosts = sum(detections_df[det_status] == DetectionStatus.DS_ALMOST)
        print('...and {} almost-positives'.format(n_almosts))


    ##%% Pull out descriptive metadata

    if options.job_name_string is not None:
        job_name_string = options.job_name_string
    else:
        # This is rare; it only happens during debugging when the caller
        # is supplying already-loaded API results.
        if options.api_output_file is None:
            job_name_string = 'unknown'
        else:
            job_name_string = os.path.basename(options.api_output_file)
    
    if options.model_version_string is not None:
        model_version_string = options.model_version_string
    else:
        
        if 'info' not in other_fields or 'detector' not in other_fields['info']:
            print('No model metadata supplied, assuming MDv4')
            model_version_string = 'MDv4 (assumed)'
        else:            
            model_version_string = other_fields['info']['detector']
    
        
    ##%% If we have ground truth, remove images we can't match to ground truth

    if ground_truth_indexed_db is not None:

        b_match = detections_df['file'].isin(
            ground_truth_indexed_db.filename_to_id)
        print(f'Confirmed filename matches to ground truth for {sum(b_match)} '
              f'of {len(detections_df)} files')

        detections_df = detections_df[b_match]
        detector_files = detections_df['file'].tolist()

        assert len(detector_files) > 0, (
            'No detection files available, possible path issue?')

        print('Trimmed detection results to {} files'.format(len(detector_files)))


    ##%% Sample images for visualization

    images_to_visualize = detections_df

    if options.num_images_to_sample is not None and options.num_images_to_sample > 0:
        images_to_visualize = images_to_visualize.sample(
            n=min(options.num_images_to_sample, len(images_to_visualize)),
            random_state=options.sample_seed)

    output_html_file = ''

    style_header = """<head>
        <style type="text/css">
        a { text-decoration: none; }
        body { font-family: segoe ui, calibri, "trebuchet ms", verdana, arial, sans-serif; }
        div.contentdiv { margin-left: 20px; }
        </style>
        </head>"""


    ##%% Fork here depending on whether or not ground truth is available

    # If we have ground truth, we'll compute precision/recall and sample tp/fp/tn/fn.
    #
    # Otherwise we'll just visualize detections/non-detections.

    if ground_truth_indexed_db is not None:

        ##%% Detection evaluation: compute precision/recall

        # numpy array of detection probabilities
        p_detection = detections_df['max_detection_conf'].values
        n_detections = len(p_detection)

        # numpy array of bools (0.0/1.0), and -1 as null value
        gt_detections = np.zeros(n_detections, dtype=float)

        for i_detection, fn in enumerate(detector_files):
            image_id = ground_truth_indexed_db.filename_to_id[fn]
            image = ground_truth_indexed_db.image_id_to_image[image_id]
            detection_status = image['_detection_status']

            if detection_status == DetectionStatus.DS_NEGATIVE:
                gt_detections[i_detection] = 0.0
            elif detection_status == DetectionStatus.DS_POSITIVE:
                gt_detections[i_detection] = 1.0
            else:
                gt_detections[i_detection] = -1.0

        # Don't include ambiguous/unknown ground truth in precision/recall analysis
        b_valid_ground_truth = gt_detections >= 0.0

        p_detection_pr = p_detection[b_valid_ground_truth]
        gt_detections_pr = (gt_detections[b_valid_ground_truth] == 1.)

        print('Including {} of {} values in p/r analysis'.format(np.sum(b_valid_ground_truth),
              len(b_valid_ground_truth)))

        precisions, recalls, thresholds = precision_recall_curve(gt_detections_pr, p_detection_pr)

        # For completeness, include the result at a confidence threshold of 1.0
        thresholds = np.append(thresholds, [1.0])

        precisions_recalls = pd.DataFrame(data={
            'confidence_threshold': thresholds,
            'precision': precisions,
            'recall': recalls
        })

        # Compute and print summary statistics
        average_precision = average_precision_score(gt_detections_pr, p_detection_pr)
        print('Average precision: {:.1%}'.format(average_precision))

        # Thresholds go up throughout precisions/recalls/thresholds; find the last
        # value where recall is at or above target.  That's our precision @ target recall.
        
        i_above_target_recall = (np.where(recalls >= options.target_recall))
        
        # np.where returns a tuple of arrays, but in this syntax where we're 
        # comparing an array with a scalar, there will only be one element.
        assert len (i_above_target_recall) == 1
        
        # Convert back to a list
        i_above_target_recall = i_above_target_recall[0].tolist()
        
        if len(i_above_target_recall) == 0:
            precision_at_target_recall = 0.0
        else:
            precision_at_target_recall = precisions[i_above_target_recall[-1]]
        print('Precision at {:.1%} recall: {:.1%}'.format(options.target_recall,
                                                          precision_at_target_recall))

        cm_predictions = np.array(p_detection_pr) > options.confidence_threshold
        cm = confusion_matrix(gt_detections_pr, cm_predictions, labels=[False,True])

        # Flatten the confusion matrix
        tn, fp, fn, tp = cm.ravel()

        precision_at_confidence_threshold = tp / (tp + fp)
        recall_at_confidence_threshold = tp / (tp + fn)
        f1 = 2.0 * (precision_at_confidence_threshold * recall_at_confidence_threshold) / \
            (precision_at_confidence_threshold + recall_at_confidence_threshold)

        print('At a confidence threshold of {:.1%}, precision={:.1%}, recall={:.1%}, f1={:.1%}'.format(
                options.confidence_threshold, precision_at_confidence_threshold,
                recall_at_confidence_threshold, f1))

        ##%% Collect classification results, if they exist

        classifier_accuracies = []

        # Mapping of classnames to idx for the confusion matrix.
        #
        # The lambda is actually kind of a hack, because we use assume that
        # the following code does not reassign classname_to_idx
        classname_to_idx = collections.defaultdict(lambda: len(classname_to_idx))

        # Confusion matrix as defaultdict of defaultdict
        #
        # Rows / first index is ground truth, columns / second index is predicted category
        classifier_cm = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))

        # iDetection = 0; fn = detector_files[iDetection]; print(fn)
        assert len(detector_files) == len(detections_df)
        for iDetection, fn in enumerate(detector_files):

            image_id = ground_truth_indexed_db.filename_to_id[fn]
            image = ground_truth_indexed_db.image_id_to_image[image_id]
            detections = detections_df['detections'].iloc[iDetection]
            pred_class_ids = [det['classifications'][0][0] \
                for det in detections if 'classifications' in det.keys()]
            pred_classnames = [classification_categories[pd] for pd in pred_class_ids]

            # If this image has classification predictions, and an unambiguous class
            # annotated, and is a positive image...
            if len(pred_classnames) > 0 \
                    and '_unambiguous_category' in image.keys() \
                    and image['_detection_status'] == DetectionStatus.DS_POSITIVE:

                # The unambiguous category, we make this a set for easier handling afterward
                gt_categories = set([image['_unambiguous_category']])
                pred_categories = set(pred_classnames)

                # Compute the accuracy as intersection of union,
                # i.e. (# of categories in both prediction and GT)
                #      divided by (# of categories in either prediction or GT
                #
                # In case of only one GT category, the result will be 1.0, if
                # prediction is one category and this category matches GT
                #
                # It is 1.0/(# of predicted top-1 categories), if the GT is
                # one of the predicted top-1 categories.
                #
                # It is 0.0, if none of the predicted categories is correct

                classifier_accuracies.append(
                    len(gt_categories & pred_categories)
                    / len(gt_categories | pred_categories)
                )
                image['_classification_accuracy'] = classifier_accuracies[-1]

                # Distribute this accuracy across all predicted categories in the
                # confusion matrix
                assert len(gt_categories) == 1
                gt_class_idx = classname_to_idx[list(gt_categories)[0]]
                for pred_category in pred_categories:
                    pred_class_idx = classname_to_idx[pred_category]
                    classifier_cm[gt_class_idx][pred_class_idx] += 1

        # ...for each file in the detection results

        # If we have classification results
        if len(classifier_accuracies) > 0:

            # Build confusion matrix as array from classifier_cm
            all_class_ids = sorted(classname_to_idx.values())
            classifier_cm_array = np.array(
                [[classifier_cm[r_idx][c_idx] for c_idx in all_class_ids] for \
                 r_idx in all_class_ids], dtype=float)
            classifier_cm_array /= (classifier_cm_array.sum(axis=1, keepdims=True) + 1e-7)

            # Print some statistics
            print('Finished computation of {} classification results'.format(
                len(classifier_accuracies)))
            print('Mean accuracy: {}'.format(np.mean(classifier_accuracies)))

            # Prepare confusion matrix output

            # Get confusion matrix as string
            sio = io.StringIO()
            np.savetxt(sio, classifier_cm_array * 100, fmt='%5.1f')
            cm_str = sio.getvalue()
            # Get fixed-size classname for each idx
            idx_to_classname = {v:k for k,v in classname_to_idx.items()}
            classname_list = [idx_to_classname[idx] for idx in sorted(classname_to_idx.values())]
            classname_headers = ['{:<5}'.format(cname[:5]) for cname in classname_list]

            # Prepend class name on each line and add to the top
            cm_str_lines = [' ' * 16 + ' '.join(classname_headers)]
            cm_str_lines += ['{:>15}'.format(cn[:15]) + ' ' + cm_line for cn, cm_line in \
                             zip(classname_list, cm_str.splitlines())]

            # Print formatted confusion matrix
            if False:
                # Actually don't, this gets really messy in all but the widest consoles
                print('Confusion matrix: ')
                print(*cm_str_lines, sep='\n')

            # Plot confusion matrix

            # To manually add more space at bottom: plt.rcParams['figure.subplot.bottom'] = 0.1
            #
            # Add 0.5 to figsize for every class. For two classes, this will result in
            # fig = plt.figure(figsize=[4,4])
            fig = plot_utils.plot_confusion_matrix(
                classifier_cm_array,
                classname_list,
                normalize=False,
                title='Confusion matrix',
                cmap=plt.cm.Blues,
                vmax=1.0,
                use_colorbar=True,
                y_label=True)
            cm_figure_relative_filename = 'confusion_matrix.png'
            cm_figure_filename = os.path.join(output_dir, cm_figure_relative_filename)
            plt.savefig(cm_figure_filename)
            plt.close(fig)

        # ...if we have classification results


        ##%% Render output

        # Write p/r table to .csv file in output directory
        pr_table_filename = os.path.join(output_dir, 'prec_recall.csv')
        precisions_recalls.to_csv(pr_table_filename, index=False)

        # Write precision/recall plot to .png file in output directory
        t = 'Precision-Recall curve: AP={:0.1%}, P@{:0.1%}={:0.1%}'.format(
            average_precision, options.target_recall, precision_at_target_recall)
        fig = plot_utils.plot_precision_recall_curve(precisions, recalls, t)
        
        pr_figure_relative_filename = 'prec_recall.png'
        pr_figure_filename = os.path.join(output_dir, pr_figure_relative_filename)
        fig.savefig(pr_figure_filename)
        plt.close(fig)


        ##%% Sampling

        # Sample true/false positives/negatives with correct/incorrect top-1
        # classification and render to html

        # Accumulate html image structs (in the format expected by write_html_image_lists)
        # for each category, e.g. 'tp', 'fp', ..., 'class_bird', ...
        images_html = collections.defaultdict(list)
        
        # Add default entries by accessing them for the first time
        [images_html[res] for res in ['tp', 'tpc', 'tpi', 'fp', 'tn', 'fn']]
        for res in images_html.keys():
            os.makedirs(os.path.join(output_dir, res), exist_ok=True)

        image_count = len(images_to_visualize)

        # Each element will be a list of 2-tuples, with elements [collection name,html info struct]
        rendering_results = []

        # Each element will be a three-tuple with elements file,max_conf,detections
        files_to_render = []

        # Assemble the information we need for rendering, so we can parallelize without
        # dealing with Pandas
        # i_row = 0; row = images_to_visualize.iloc[0]
        for _, row in images_to_visualize.iterrows():

            # Filenames should already have been normalized to either '/' or '\'
            files_to_render.append([row['file'], row['max_detection_conf'], row['detections']])

        start_time = time.time()
        if options.parallelize_rendering:
            if options.parallelize_rendering_n_cores is None:                
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool()
                else:
                    pool = Pool()
            else:
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool(options.parallelize_rendering_n_cores)
                    worker_string = 'threads'
                else:
                    pool = Pool(options.parallelize_rendering_n_cores)
                    worker_string = 'processes'
                print('Rendering images with {} {}'.format(options.parallelize_rendering_n_cores,
                                                           worker_string))
                
            rendering_results = list(tqdm(pool.imap(
                partial(render_image_with_gt,
                        ground_truth_indexed_db=ground_truth_indexed_db,
                        detection_categories=detection_categories,
                        classification_categories=classification_categories,
                        options=options), 
                files_to_render), total=len(files_to_render)))
        else:
            for file_info in tqdm(files_to_render):
                rendering_results.append(render_image_with_gt(
                    file_info,ground_truth_indexed_db,
                    detection_categories,classification_categories))
        elapsed = time.time() - start_time

        # Map all the rendering results in the list rendering_results into the
        # dictionary images_html
        image_rendered_count = 0
        for rendering_result in rendering_results:
            if rendering_result is None:
                continue
            image_rendered_count += 1
            for assignment in rendering_result:
                images_html[assignment[0]].append(assignment[1])

        # Prepare the individual html image files
        image_counts = prepare_html_subpages(images_html, output_dir, options)

        print('{} images rendered (of {})'.format(image_rendered_count,image_count))

        # Write index.html
        all_tp_count = image_counts['tp'] + image_counts['tpc'] + image_counts['tpi']
        total_count = all_tp_count + image_counts['tn'] + image_counts['fp'] + image_counts['fn']

        classification_detection_results = """&nbsp;&nbsp;&nbsp;&nbsp;<a href="tpc.html">with all correct top-1 predictions (TPC)</a> ({})<br/>
           &nbsp;&nbsp;&nbsp;&nbsp;<a href="tpi.html">with one or more incorrect top-1 prediction (TPI)</a> ({})<br/>
           &nbsp;&nbsp;&nbsp;&nbsp;<a href="tp.html">without classification evaluation</a><sup>*</sup> ({})<br/>""".format(
            image_counts['tpc'],
            image_counts['tpi'],
            image_counts['tp']
        )

        index_page = """<html>
        {}
        <body>
        <h2>Evaluation</h2>

        <h3>Job metadata</h3>
        
        <div class="contentdiv">
        <p>Job name: {}<br/>
        <p>Model version: {}</p>
        </div>
        
        <h3>Sample images</h3>
        <div class="contentdiv">
        <p>A sample of {} images, annotated with detections above {:.1%} confidence.</p>
        <a href="tp.html">True positives (TP)</a> ({}) ({:0.1%})<br/>
        CLASSIFICATION_PLACEHOLDER_1
        <a href="tn.html">True negatives (TN)</a> ({}) ({:0.1%})<br/>
        <a href="fp.html">False positives (FP)</a> ({}) ({:0.1%})<br/>
        <a href="fn.html">False negatives (FN)</a> ({}) ({:0.1%})<br/>
        CLASSIFICATION_PLACEHOLDER_2
        </div>
        """.format(
            style_header,job_name_string,model_version_string,
            image_count, options.confidence_threshold,
            all_tp_count, all_tp_count/total_count,
            image_counts['tn'], image_counts['tn']/total_count,
            image_counts['fp'], image_counts['fp']/total_count,
            image_counts['fn'], image_counts['fn']/total_count
        )

        index_page += """
            <h3>Detection results</h3>
            <div class="contentdiv">
            <p>At a confidence threshold of {:0.1%}, precision={:0.1%}, recall={:0.1%}</p>
            <p><strong>Precision/recall summary for all {} images</strong></p><img src="{}"><br/>
            </div>
            """.format(
                options.confidence_threshold, precision_at_confidence_threshold, recall_at_confidence_threshold,
                len(detections_df), pr_figure_relative_filename
           )

        if len(classifier_accuracies) > 0:
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_1',classification_detection_results)
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_2',"""<p><sup>*</sup>We do not evaluate the classification result of images
                if the classification information is missing, if the image contains
                categories like &lsquo;empty&rsquo; or &lsquo;human&rsquo;, or if the image has multiple
                classification labels.</p>""")
        else:
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_1','')
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_2','')

        if len(classifier_accuracies) > 0:
            index_page += """
                <h3>Classification results</h3>
                <div class="contentdiv">
                <p>Classification accuracy: {:.2%}<br>
                The accuracy is computed only for images with exactly one classification label.
                The accuracy of an image is computed as 1/(number of unique detected top-1 classes),
                i.e. if the model detects multiple boxes with different top-1 classes, then the accuracy
                decreases and the image is put into 'TPI'.</p>
                <p>Confusion matrix:</p>
                <p><img src="{}"></p>
                <div style='font-family:monospace;display:block;'>{}</div>
                </div>
                """.format(
                    np.mean(classifier_accuracies),
                    cm_figure_relative_filename,
                    "<br>".join(cm_str_lines).replace(' ', '&nbsp;')
                )

        # Show links to each GT class
        #
        # We could do this without classification results; currently we don't.
        if len(classname_to_idx) > 0:

            index_page += '<h3>Images of specific classes</h3><br/><div class="contentdiv">'
            # Add links to all available classes
            for cname in sorted(classname_to_idx.keys()):
                index_page += '<a href="class_{0}.html">{0}</a> ({1})<br>'.format(
                    cname,
                    len(images_html['class_{}'.format(cname)]))
            index_page += '</div>'

        # Close body and html tags
        index_page += '</body></html>'
        output_html_file = os.path.join(output_dir, 'index.html')
        with open(output_html_file, 'w') as f:
            f.write(index_page)

        print('Finished writing html to {}'.format(output_html_file))

    # ...for each image


    ##%% Otherwise, if we don't have ground truth...

    else:

        ##%% Sample detections/non-detections

        # Accumulate html image structs (in the format expected by write_html_image_list)
        # for each category
        images_html = collections.defaultdict(list)
        images_html['non_detections']

        # Add default entries by accessing them for the first time

        # Maps detection categories - e.g. "human" - to result set names, e.g.
        # "detections_human"
        detection_categories_to_results_name = {}

        if not options.separate_detections_by_category:
            images_html['detections']
        else:
            # Add a set of results for each category and combination of categories
            keys = detection_categories.keys()
            subsets = []
            for L in range(1, len(keys)+1):
                for subset in itertools.combinations(keys, L):
                    subsets.append(subset)
            for subset in subsets:
                sorted_subset = tuple(sorted(subset))
                results_name = 'detections'
                for category_id in sorted_subset:
                    results_name = results_name + '_' + detection_categories[category_id]
                images_html[results_name]
                detection_categories_to_results_name[sorted_subset] = results_name

        if options.include_almost_detections:
            images_html['almost_detections']

        # Create output directories
        for res in images_html.keys():
            os.makedirs(os.path.join(output_dir, res), exist_ok=True)

        image_count = len(images_to_visualize)
        has_classification_info = False

        # Each element will be a list of 2-tuples, with elements [collection name,html info struct]
        rendering_results = []

        # list of 3-tuples with elements (file, max_conf, detections)
        files_to_render = []

        # Assemble the information we need for rendering, so we can parallelize without
        # dealing with Pandas
        # i_row = 0; row = images_to_visualize.iloc[0]
        for _, row in images_to_visualize.iterrows():

            assert isinstance(row['detections'],list)
            
            # Filenames should already have been normalized to either '/' or '\'
            files_to_render.append([row['file'],
                                    row['max_detection_conf'],
                                    row['detections']])

        start_time = time.time()
        if options.parallelize_rendering:
            
            if options.parallelize_rendering_n_cores is None:                
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool()
                else:
                    pool = Pool()
            else:
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool(options.parallelize_rendering_n_cores)
                    worker_string = 'threads'
                else:
                    pool = Pool(options.parallelize_rendering_n_cores)
                    worker_string = 'processes'
                print('Rendering images with {} {}'.format(options.parallelize_rendering_n_cores,
                                                           worker_string))
                
            # render_image_no_gt(file_info,detection_categories_to_results_name,
            # detection_categories,classification_categories)

            rendering_results = list(tqdm(pool.imap(
                partial(render_image_no_gt, 
                        detection_categories_to_results_name=detection_categories_to_results_name,
                        detection_categories=detection_categories,
                        classification_categories=classification_categories,
                        options=options),
                        files_to_render), total=len(files_to_render)))
        else:
            for file_info in tqdm(files_to_render):
                rendering_results.append(render_image_no_gt(file_info,
                                                            detection_categories_to_results_name,
                                                            detection_categories,
                                                            classification_categories,
                                                            options=options))
                
        elapsed = time.time() - start_time

        # Map all the rendering results in the list rendering_results into the
        # dictionary images_html
        image_rendered_count = 0
        for rendering_result in rendering_results:
            if rendering_result is None:
                continue
            image_rendered_count += 1
            for assignment in rendering_result:
                if 'class' in assignment[0]:
                    has_classification_info = True
                images_html[assignment[0]].append(assignment[1])

        # Prepare the individual html image files
        image_counts = prepare_html_subpages(images_html, output_dir, options)

        if image_rendered_count == 0:
            seconds_per_image = 0.0
        else:
            seconds_per_image = elapsed/image_rendered_count

        print('Rendered {} images (of {}) in {} ({} per image)'.format(image_rendered_count,
              image_count,humanfriendly.format_timespan(elapsed),
              humanfriendly.format_timespan(seconds_per_image)))

        # Write index.html

        # We can't just sum these, because image_counts includes images in both their
        # detection and classification classes
        # total_images = sum(image_counts.values())
        total_images = 0
        for k in image_counts.keys():
            v = image_counts[k]
            if has_classification_info and k.startswith('class_'):
                continue
            total_images += v

        if total_images != image_count:
            print('Warning, missing images: image_count is {}, total_images is {}'.format(total_images,image_count))
        
        almost_detection_string = ''
        if options.include_almost_detections:
            almost_detection_string = ' (&ldquo;almost detection&rdquo; threshold at {:.1%})'.format(
                options.almost_detection_confidence_threshold)

        index_page = """<html>\n{}\n<body>\n
        <h2>Visualization of results for {}</h2>\n
        <p>A sample of {} images (of {} total)FAILURE_PLACEHOLDER, annotated with detections above {:.1%} confidence{}.</p>\n
        
        <div class="contentdiv">
        <p>Model version: {}</p>
        </div>
        
        <h3>Sample images</h3>\n
        <div class="contentdiv">\n""".format(
            style_header, job_name_string, image_count, len(detections_df), options.confidence_threshold,
            almost_detection_string, model_version_string)

        failure_string = ''
        if n_failures is not None:
            failure_string = ' ({} failures)'.format(n_failures)        
        index_page = index_page.replace('FAILURE_PLACEHOLDER',failure_string)
        
        def result_set_name_to_friendly_name(result_set_name):
            friendly_name = ''
            friendly_name = result_set_name.replace('_','-')
            if friendly_name.startswith('detections-'):
                friendly_name = friendly_name.replace('detections-', 'detections: ')
            friendly_name = friendly_name.capitalize()
            return friendly_name

        for result_set_name in images_html.keys():

            # Don't print classification classes here; we'll do that later with a slightly
            # different structure
            if has_classification_info and result_set_name.lower().startswith('class_'):
                continue

            filename = result_set_name + '.html'
            label = result_set_name_to_friendly_name(result_set_name)
            image_count = image_counts[result_set_name]
            if total_images == 0:
                image_fraction = -1
            else:
                image_fraction = image_count / total_images
            index_page += '<a href="{}">{}</a> ({}, {:.1%})<br/>\n'.format(
                filename,label,image_count,image_fraction)

        index_page += '</div>\n'

        if has_classification_info:
            index_page += '<h3>Images of detected classes</h3>'
            index_page += '<p>The same image might appear under multiple classes if multiple species were detected.</p>\n'
            index_page += '<p>Classifications with confidence less than {:.1%} confidence are considered "unreliable".</p>\n'.format(
                options.classification_confidence_threshold)
            index_page += '<div class="contentdiv">\n'

            # Add links to all available classes
            class_names = sorted(classification_categories.values())
            if 'class_unreliable' in images_html.keys():
                class_names.append('unreliable')

            if options.sort_classification_results_by_count:
                class_name_to_count = {}                
                for cname in class_names:
                    ccount = len(images_html['class_{}'.format(cname)])
                    class_name_to_count[cname] = ccount
                class_names = sorted(class_names,key=lambda x: class_name_to_count[x],reverse=True)                
                    
            for cname in class_names:
                ccount = len(images_html['class_{}'.format(cname)])
                if ccount > 0:
                    index_page += '<a href="class_{}.html">{}</a> ({})<br/>\n'.format(
                        cname, cname.lower(), ccount)
            index_page += '</div>\n'

        index_page += '</body></html>'
        output_html_file = os.path.join(output_dir, 'index.html')
        with open(output_html_file, 'w') as f:
            f.write(index_page)

        print('Finished writing html to {}'.format(output_html_file))

        # os.startfile(output_html_file)

    # ...if we do/don't have ground truth

    ppresults.output_html_file = output_html_file
    return ppresults

# ...process_batch_results


#%% Interactive driver(s)

if False:

    #%%

    base_dir = r'G:\temp\md'
    options = PostProcessingOptions()
    options.image_base_dir = base_dir
    options.output_dir = os.path.join(base_dir, 'postprocessing')
    options.api_output_filename_replacements = {} # {'20190430cameratraps\\':''}
    options.ground_truth_filename_replacements = {} # {'\\data\\blob\\':''}
    options.api_output_file = os.path.join(base_dir, 'results.json')
    options.ground_truth_json_file = os.path.join(base_dir, 'gt.json')
    # options.unlabeled_classes = ['human']

    ppresults = process_batch_results(options)
    # os.start(ppresults.output_html_file)


#%% Command-line driver

def main():
    
    options = PostProcessingOptions()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'api_output_file',
        help='path to .json file produced by the batch inference API')
    parser.add_argument(
        'output_dir',
        help='base directory for output')
    parser.add_argument(
        '--image_base_dir', default=options.image_base_dir,
        help='base directory for images (optional, can compute statistics '
             'without images)')
    parser.add_argument(
        '--ground_truth_json_file', default=options.ground_truth_json_file,
        help='ground truth labels (optional, can render detections without '
             'ground truth), in the COCO Camera Traps format')
    parser.add_argument(
        '--confidence_threshold', type=float,
        default=options.confidence_threshold,
        help='Confidence threshold for statistics and visualization')
    parser.add_argument(
        '--almost_detection_confidence_threshold', type=float,
        default=options.almost_detection_confidence_threshold,
        help='Almost-detection confidence threshold for statistics and visualization')
    parser.add_argument(
        '--target_recall', type=float, default=options.target_recall,
        help='Target recall (for statistics only)')
    parser.add_argument(
        '--num_images_to_sample', type=int,
        default=options.num_images_to_sample,
        help='number of images to visualize, -1 for all images (default: 500)')
    parser.add_argument(
        '--viz_target_width', type=int, default=options.viz_target_width,
        help='Output image width')
    parser.add_argument(
        '--include_almost_detections', action='store_true',
        help='Include a separate category for images just above a second confidence threshold')
    parser.add_argument(
        '--random_output_sort', action='store_true',
        help='Sort output randomly (defaults to sorting by filename)')
    parser.add_argument(
        '--n_cores', type=int, default=1,
        help='Number of threads to use for rendering (default: 1)')
    parser.add_argument(
        '--parallelize_rendering_with_processes', 
        action='store_true',
        help='Should we use processes (instead of threads) for parallelization?')
    parser.add_argument(
        '--no_separate_detections_by_category', 
        action='store_true',
        help='Collapse all categories into just "detections" and "non-detections"')    
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    args.sort_html_by_filename = (not args.random_output_sort)
    if args.n_cores != 1:
        assert (args.n_cores > 1), 'Illegal number of cores: {}'.format(args.n_cores)
        if args.parallelize_rendering_with_processes:
            args.parallelize_rendering_with_threads = False
        args.parallelize_rendering = True
        args.parallelize_rendering_n_cores = args.n_cores        

    args_to_object(args, options)

    if args.no_separate_detections_by_category:
        options.separate_detections_by_category = False
    
    process_batch_results(options)


if __name__ == '__main__':
    main()



########
#
# prepare_lila_taxonomy_release.py
#
# Given the private intermediate taxonomy mapping (produced by map_new_lila_datasets.py), 
# prepare the public (release) taxonomy mapping file.
#
########

#%% Imports and constants

import os
import json
import pandas as pd

lila_taxonomy_file = os.path.expanduser('~/git/agentmorrisprivate/lila-taxonomy/lila-taxonomy-mapping.csv')
release_taxonomy_file = os.path.expanduser('~/lila/lila-taxonomy-mapping_release.22.08.22.0000.csv')

# Created by get_lila_category_list.py... contains counts for each category
lila_dataset_to_categories_file = os.path.expanduser('~/lila/lila_categories_list/lila_dataset_to_categories.json')

assert os.path.isfile(lila_dataset_to_categories_file)
assert os.path.isfile(lila_taxonomy_file)


#%% Find out which categories are actually used

df = pd.read_csv(lila_taxonomy_file)

with open(lila_dataset_to_categories_file,'r') as f:
    lila_dataset_to_categories = json.load(f)

used_category_mappings = []

# dataset_name = datasets_to_map[0]
for dataset_name in lila_dataset_to_categories.keys():
    
    ds_categories = lila_dataset_to_categories[dataset_name]
    for category in ds_categories:
        category_name = category['name'].lower()
        assert ':' not in category_name
        mapping_name = dataset_name + ':' + category_name
        used_category_mappings.append(mapping_name)

df['used'] = False

# i_row = 0; row = df.iloc[i_row]; row
for i_row,row in df.iterrows():
    ds_name = row['dataset_name']
    query = row['query']
    mapping_name = ds_name + ':' + query
    if mapping_name in used_category_mappings:
        df.loc[i_row,'used'] = True
    else:
        print('Dropping unused mapping {}'.format(mapping_name))

df = df[df.used]
df = df.drop('used',axis=1)


#%% Generate the final output file

assert not os.path.isfile(release_taxonomy_file)

known_levels = ['stateofmatter',
                     'kingdom',
                     'phylum','subphylum',
                     'superclass','class','subclass','infraclass',
                     'superorder','order','parvorder','suborder','infraorder',
                     'zoosection',
                     'superfamily','family','subfamily','tribe',
                     'genus',
                     'species','subspecies','variety']

levels_to_include = ['kingdom',
                     'phylum','subphylum',
                     'superclass','class','subclass','infraclass',
                     'superorder','order','suborder','infraorder',
                     'superfamily','family','subfamily','tribe',
                     'genus',
                     'species','subspecies','variety']

levels_to_exclude = ['stateofmatter','zoosection','parvorder']

for s in levels_to_exclude:
    assert s not in levels_to_include
    
levels_used = set()

# i_row = 0; row = df.iloc[i_row]; row
for i_row,row in df.iterrows():
    
    if not isinstance(row['scientific_name'],str):
        assert not isinstance(row['taxonomy_string'],str)
        continue
    
    taxonomic_match = eval(row['taxonomy_string'])
    
    # match_at_level = taxonomic_match[0]
    for match_at_level in taxonomic_match:
        assert len(match_at_level) == 4
        levels_used.add(match_at_level[1])
        
levels_used = [s for s in levels_used if isinstance(s,str)]

for s in levels_used:
    assert s in levels_to_exclude or s in levels_to_include, 'Unrecognized level {}'.format(s)

for s in levels_to_include:
    assert s in levels_used
    
for s in levels_to_include:
    df[s] = ''
    
# i_row = 0; row = df.iloc[i_row]; row
for i_row,row in df.iterrows():
    
    if not isinstance(row['scientific_name'],str):
        assert not isinstance(row['taxonomy_string'],str)
        continue
    
    # E.g.: (43117, 'genus', 'lepus', ['hares and jackrabbits']
    taxonomic_match = eval(row['taxonomy_string'])
    
    for match_at_level in taxonomic_match:
        level = match_at_level[1]
        if level in levels_to_include:
            df.loc[i_row,level] = match_at_level[2]

df = df.drop('source',axis=1)
df.to_csv(release_taxonomy_file,header=True,index=False)



########
#
# get_image_sizes.py
#
# Given a json-formatted list of image filenames, retrieve the width and height of every image.
#
########

#%% Constants and imports

import argparse
import json
import os
from PIL import Image
import sys

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial
from tqdm import tqdm

image_base = ''
default_n_threads = 1
use_threads = False


#%% Processing functions

def process_image(image_path,image_prefix=None):
    
    if image_prefix is not None:
        full_path = os.path.join(image_prefix,image_path)
    else:
        full_path = image_path
    
    # Is this image on disk?
    if not os.path.isfile(full_path):
        print('Could not find image {}'.format(full_path))
        return (image_path,-1,-1)

    try:        
        pil_im = Image.open(full_path)
        w = pil_im.width            
        h = pil_im.height
        return (image_path,w,h)
    except Exception as e:    
        print('Error reading image {}: {}'.format(full_path,str(e)))
        return (image_path,-1,-1)
    
    
def process_images(filenames,image_prefix=None,n_threads=default_n_threads):
    
    if n_threads <= 1:
        
        all_results = []
        for i_file,fn in tqdm(enumerate(filenames),total=len(filenames)):
            all_results.append(process_image(fn,image_prefix=image_prefix))
    
    else:
        
        print('Creating a pool with {} threads'.format(n_threads))
        if use_threads:
            pool = ThreadPool(n_threads)        
        else:
            pool = Pool(n_threads)
        # all_results = list(tqdm(pool.imap(process_image, filenames), total=len(filenames)))
        all_results = list(tqdm(pool.imap(
            partial(process_image,image_prefix=image_prefix), filenames), total=len(filenames)))
                
    return all_results


def process_list_file(input_file,output_file,image_prefix=None,n_threads=default_n_threads):
    
    assert os.path.isdir(os.path.dirname(output_file))
    assert os.path.isfile(input_file)
    
    with open(input_file,'r') as f:        
        filenames = json.load(f)
    filenames = [s.strip() for s in filenames]
    
    all_results = process_images(filenames,image_prefix=image_prefix,n_threads=n_threads)
    
    with open(output_file,'w') as f:
        json.dump(all_results,f,indent=2)
    
    
#%% Interactive driver

if False:

    pass    

    #%%
    
    # List images in a test folder
    base_dir = r'c:\temp\test_images'
    image_list_file = os.path.join(base_dir,'images.json')
    relative_image_list_file = os.path.join(base_dir,'images_relative.json')
    image_size_file = os.path.join(base_dir,'image_sizes.json')
    from md_utils import path_utils
    image_names = path_utils.find_images(base_dir,recursive=True)
    
    with open(image_list_file,'w') as f:
        json.dump(image_names,f,indent=2)
        
    relative_image_names = []
    for s in image_names:
        relative_image_names.append(os.path.relpath(s,base_dir))
    
    with open(relative_image_list_file,'w') as f:
        json.dump(relative_image_names,f,indent=2)
    
    
    #%%
    
    # process_list_file(image_list_file,image_size_file,image_prefix=base_dir)
    process_list_file(relative_image_list_file,image_size_file,image_prefix=base_dir,n_threads=4)
    
    
#%% Command-line driver
    
def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file',type=str)
    parser.add_argument('output_file',type=str)
    parser.add_argument('--image_prefix', type=str, default=None)
    parser.add_argument('--n_threads', type=int, default=default_n_threads)
                        
    if len(sys.argv[1:])==0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()
    
    process_list_file(args.input_file,args.output_file,args.image_prefix,args.n_threads)
    

if __name__ == '__main__':
    
    main()



########
#
# top_folders_to_bottom.py
#
# Given a base folder with files like:
#
# A/1/2/a.jpg
# B/3/4/b.jpg
#
# ...moves the top-level folders to the bottom in a new output folder, i.e., creates:
#
# 1/2/A/a.jpg
# 3/4/B/b.jpg
#
# In practice, this is used to make this:
# 
# animal/camera01/image01.jpg
#
# ...look like:
#
# camera01/animal/image01.jpg
#
########

#%% Constants and imports

import os
import sys
import shutil
import argparse

from pathlib import Path
from tqdm import tqdm

from functools import partial
from multiprocessing.pool import ThreadPool

class TopFoldersToBottomOptions:
    
    def __init__(self,input_folder,output_folder,copy=True,n_threads=1):
        self.copy = copy
        self.n_threads = n_threads
        self.input_folder = input_folder
        self.output_folder = output_folder
        self.overwrite = False
        
        
#%% Support functions

def path_is_abs(p): return (len(p) > 1) and (p[0] == '/' or p[1] == ':')


#%% Main functions

def process_file(relative_filename,options,execute=True):
    
    assert ('/' in relative_filename) and ('\\' not in relative_filename) and (not path_is_abs(relative_filename))
    
    # Find top-level folder
    tokens = relative_filename.split('/')
    top_level_folder = tokens.pop(0)
    tokens.insert(len(tokens)-1,top_level_folder)
    
    # Find file/folder names
    output_relative_path = '/'.join(tokens)
    output_relative_folder = '/'.join(tokens[0:-1])
    
    output_absolute_folder = os.path.join(options.output_folder,output_relative_folder)
    output_absolute_path = os.path.join(options.output_folder,output_relative_path)

    if execute:
        
        os.makedirs(output_absolute_folder,exist_ok=True)
        
        input_absolute_path = os.path.join(options.input_folder,relative_filename)
        
        if not options.overwrite:
            assert not os.path.isfile(output_absolute_path), \
                'Error: output file {} exists'.format(output_absolute_path)
            
        # Move or copy
        if options.copy:
            shutil.copy(input_absolute_path, output_absolute_path)
        else:
            shutil.move(input_absolute_path, output_absolute_path)

    return output_absolute_path
    
# ...def process_file()


def top_folders_to_bottom(options):
    
    os.makedirs(options.output_folder,exist_ok=True)
    
    # Enumerate input folder
    print('Enumerating files...')
    files = list(Path(options.input_folder).rglob('*'))
    files = [p for p in files if not p.is_dir()]
    files = [str(s) for s in files]
    print('Enumerated {} files'.format(len(files)))
    
    # Convert absolute paths to relative paths
    relative_files = [os.path.relpath(s,options.input_folder) for s in files]
    
    # Standardize delimiters
    relative_files = [s.replace('\\','/') for s in relative_files]
    
    base_files = [s for s in relative_files if '/' not in s]
    if len(base_files) > 0:
        print('Warning: ignoring {} files in the base folder'.format(len(base_files)))
        relative_files = [s for s in relative_files if '/' in s]
    
    # Make sure each input file maps to a unique output file
    absolute_output_files = [process_file(s, options, execute=False) for s in relative_files]
    assert len(absolute_output_files) == len(set(absolute_output_files)),\
        "Error: input filenames don't map to unique output filenames"
        
    # relative_filename = relative_files[0]
    
    # Loop
    if options.n_threads <= 1:
        
        for relative_filename in tqdm(relative_files):
            process_file(relative_filename,options)
    
    else:
        
        print('Starting a pool with {} threads'.format(options.n_threads))
        pool = ThreadPool(options.n_threads)
        process_file_with_options = partial(process_file, options=options)
        _ = list(tqdm(pool.imap(process_file_with_options, relative_files), total=len(relative_files)))

# ...def top_folders_to_bottom()        
        

#%% Interactive driver
        
if False:

    pass

    #%%
    
    input_folder = r"G:\temp\output"
    output_folder = r"G:\temp\output-inverted"    
    options = TopFoldersToBottomOptions(input_folder,output_folder,copy=True,n_threads=10)
    
    #%%
    
    top_folders_to_bottom(options)


#%% Command-line driver   

# python top_folders_to_bottom.py "g:\temp\separated_images" "g:\temp\separated_images_inverted" --n_threads 100

def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_folder', type=str, help='Input image folder')
    parser.add_argument('output_folder', type=str, help='Output image folder')

    parser.add_argument('--copy', action='store_true', 
                        help='Copy images, instead of moving (moving is the default)')
    parser.add_argument('--overwrite', action='store_true', 
                        help='Allow image overwrite (default=False)')
    parser.add_argument('--n_threads', type=int, default=1,
                        help='Number of threads to use for parallel operation (default=1)')
    
    if len(sys.argv[1:])==0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()    
    
    # Convert to an options object
    options = TopFoldersToBottomOptions(
        args.input_folder,args.output_folder,copy=args.copy,n_threads=args.n_threads)
    
    top_folders_to_bottom(options)
    
    
if __name__ == '__main__':
    
    main()
    

    


########
#
# cct_to_wi.py
#
# Converts COCO Camera Traps .json files to the Wildlife Insights
# batch upload format
#
# Also see:
#
# https://github.com/ConservationInternational/Wildlife-Insights----Data-Migration
#
# https://data.naturalsciences.org/wildlife-insights/taxonomy/search
#
########

#%% Imports

import os
import json
import pandas as pd
from collections import defaultdict 


#%% Paths

# A COCO camera traps file with information about this dataset
input_file = r'c:\temp\camera_trap_images_no_people\bellevue_camera_traps.2020-12-26.json'
assert os.path.isfile(input_file)

# A .json dictionary mapping common names in this dataset to dictionaries with the 
# WI taxonomy fields: common_name, wi_taxon_id, class, orer, family, genus, species
taxonomy_file = r'c:\temp\camera_trap_images_no_people\belleve_camera_traps_to_wi.json'
assert os.path.isfile(taxonomy_file)

templates_dir = r'c:\temp\wi_batch_upload_templates'
assert os.path.isdir(templates_dir)

output_base = r'c:\temp\wi_output'
os.makedirs(output_base,exist_ok = True)

    
#%% Constants

projects_file_name = 'Template Wildlife Insights Batch Upload - Projectv1.0.csv'
deployments_file_name = 'Template Wildlife Insights Batch Upload - Deploymentv1.0.csv'
images_file_name = 'Template Wildlife Insights Batch Upload - Imagev1.0.csv'
cameras_file_name = 'Template Wildlife Insights Batch Upload - Camerav1.0.csv'

assert all([os.path.isfile(os.path.join(templates_dir,fn)) for fn in \
            [projects_file_name,deployments_file_name,images_file_name,cameras_file_name]])


#%% Project information

project_info = {}
project_info['project_name'] = 'Bellevue Camera Traps'
project_info['project_id'] = 'bct_001'
project_info['project_short_name'] = 'BCT'
project_info['project_objectives'] = 'none'
project_info['project_species'] = 'Multiple'
project_info['project_species_individual'] = ''
project_info['project_sensor_layout'] = 'Convenience'
project_info['project_sensor_layout_targeted_type'] = ''
project_info['project_bait_use'] = 'No'
project_info['project_bait_type'] = 'None'
project_info['project_stratification'] = 'No'
project_info['project_stratification_type'] = ''
project_info['project_sensor_method'] = 'Sensor Detection'
project_info['project_individual_animals'] = 'No'
project_info['project_admin'] = 'Dan Morris'
project_info['project_admin_email'] = 'cameratraps@lila.science'
project_info['country_code'] = 'USA'
project_info['embargo'] = str(0)
project_info['initiative_id'] = ''
project_info['metadata_license'] = 'CC0'
project_info['image_license'] = 'CC0'

project_info['project_blank_images'] = 'No'
project_info['project_sensor_cluster'] = 'No'

camera_info = {}
camera_info['project_id'] = project_info['project_id'] 
camera_info['camera_id'] = '0000'
camera_info['make'] = ''
camera_info['model'] = ''
camera_info['serial_number'] = ''
camera_info['year_purchased'] = ''

deployment_info = {}

deployment_info['project_id'] = project_info['project_id'] 
deployment_info['deployment_id'] = 'test_deployment'
deployment_info['subproject_name'] = 'test_subproject'
deployment_info['subproject_design'] = ''
deployment_info['placename'] = 'yard'
deployment_info['longitude'] = '47.6101'
deployment_info['latitude'] = '-122.2015'
deployment_info['start_date'] = '2016-01-01 00:00:00'
deployment_info['end_date'] = '2026-01-01 00:00:00'
deployment_info['event_name'] = ''
deployment_info['event_description'] = ''
deployment_info['event_type'] = ''
deployment_info['bait_type'] = ''
deployment_info['bait_description'] = ''
deployment_info['feature_type'] = 'None'
deployment_info['feature_type_methodology'] = ''
deployment_info['camera_id'] = camera_info['camera_id']
deployment_info['quiet_period'] = str(60)
deployment_info['camera_functioning'] = 'Camera Functioning'
deployment_info['sensor_height'] = 'Chest height'
deployment_info['height_other'] = ''
deployment_info['sensor_orientation'] = 'Parallel'
deployment_info['orientation_other'] = ''
deployment_info['recorded_by'] = 'Dan Morris'

image_info = {}
image_info['identified_by'] = 'Dan Morris'


#%% Read templates

def parse_fields(templates_dir,file_name):
    
    with open(os.path.join(templates_dir,file_name),'r') as f:
        lines = f.readlines()
        lines = [s.strip() for s in lines if len(s.strip().replace(',','')) > 0]
        assert len(lines) == 1, 'Error processing template {}'.format(file_name)
        fields = lines[0].split(',')
        print('Parsed {} columns from {}'.format(len(fields),file_name))
    return fields

projects_fields = parse_fields(templates_dir,projects_file_name)
deployments_fields = parse_fields(templates_dir,deployments_file_name)
images_fields = parse_fields(templates_dir,images_file_name)
cameras_fields = parse_fields(templates_dir,cameras_file_name)


#%% Compare dictionary to template lists

def compare_info_to_template(info,template_fields,name):
    
    for s in info.keys():
        assert s in template_fields,'Field {} not specified in {}_fields'.format(s,name)
    for s in template_fields:
        assert s in info.keys(),'Field {} not specified in {}_info'.format(s,name)


def write_table(file_name,info,template_fields):
    
    assert len(info) == len(template_fields)
    
    project_output_file = os.path.join(output_base,file_name)
    with open(project_output_file,'w') as f:
    
        # Write the header
        for i_field,s in enumerate(template_fields):
            f.write(s)
            if i_field != len(template_fields)-1:
                f.write(',')
        f.write('\n')
        
        # Write values
        for i_field,s in enumerate(template_fields):
            f.write(info[s])
            if i_field != len(template_fields)-1:
                f.write(',')
        f.write('\n')
    

#%% Project file

compare_info_to_template(project_info,projects_fields,'project')
write_table(projects_file_name,project_info,projects_fields)


#%% Camera file

compare_info_to_template(camera_info,cameras_fields,'camera')
write_table(cameras_file_name,camera_info,cameras_fields)


#%% Deployment file

compare_info_to_template(deployment_info,deployments_fields,'deployment')
write_table(deployments_file_name,deployment_info,deployments_fields)


#%% Images file

# Read .json file with image information
with open(input_file,'r') as f:
    input_data = json.load(f)

# Read taxonomy dictionary
with open(taxonomy_file,'r') as f:
    taxonomy_mapping = json.load(f)
    
url_base = taxonomy_mapping['url_base']
taxonomy_mapping = taxonomy_mapping['taxonomy']

# Populate output information
# df = pd.DataFrame(columns = images_fields)

category_id_to_name = {cat['id']:cat['name'] for cat in input_data['categories']}

image_id_to_annotations = defaultdict(list)

annotations = input_data['annotations']
                         
# annotation = annotations[0]
for annotation in annotations:
    image_id_to_annotations[annotation['image_id']].append(
        category_id_to_name[annotation['category_id']])

rows = []

# im = input_data['images'][0]
for im in input_data['images']:

    row = {}
    
    url = url_base + im['file_name'].replace('\\','/')
    row['project_id'] = project_info['project_id']
    row['deployment_id'] = deployment_info['deployment_id']
    row['image_id'] = im['id']
    row['location'] = url
    row['identified_by'] = image_info['identified_by']
    
    category_names = image_id_to_annotations[im['id']]
    assert len(category_names) == 1
    category_name = category_names[0]
    
    taxon_info = taxonomy_mapping[category_name]
    
    assert len(taxon_info.keys()) == 7
    
    for s in taxon_info.keys():
        row[s] = taxon_info[s]    
    
    # We don't have counts, but we can differentiate between zero and 1
    if category_name == 'empty':
        row['number_of_objects'] = 0
    else:
        row['number_of_objects'] = 1
        
    row['uncertainty'] = None
    row['timestamp'] = im['datetime']; assert isinstance(im['datetime'],str)
    row['highlighted'] = 0
    row['age'] = None
    row['sex'] = None
    row['animal_recognizable'] = 'No'
    row['individual_id'] = None
    row['individual_animal_notes'] = None
    row['markings'] = None
    
    assert len(row) == len(images_fields)
    rows.append(row)
    
df = pd.DataFrame(rows)

df.to_csv(os.path.join(output_base,images_file_name),index=False)



########
#
# run_detector.py
#
# Module to run an animal detection model on images.
# 
# The main function in this script also renders the predicted
# bounding boxes on images and saves the resulting images (with bounding boxes).
# 
# This script is not a good way to process lots of images (tens of thousands,
# say). It does not facilitate checkpointing the results so if it crashes you
# would have to start from scratch. If you want to run a detector (e.g., ours)
# on lots of images, you should check out:
# 
# 1) run_detector_batch.py (for local execution)
# 
# 2) https://github.com/agentmorris/MegaDetector/tree/master/api/batch_processing
#    (for running large jobs on Azure ML)
# 
# To run this script, we recommend you set up a conda virtual environment
# following instructions in the Installation section on the main README, using
# `environment-detector.yml` as the environment file where asked.
# 
# This is a good way to test our detector on a handful of images and get
# super-satisfying, graphical results.  It's also a good way to see how fast a
# detector model will run on a particular machine.
# 
# If you would like to *not* use the GPU on the machine, set the environment
# variable CUDA_VISIBLE_DEVICES to "-1".
# 
# If no output directory is specified, writes detections for c:\foo\bar.jpg to
# c:\foo\bar_detections.jpg.
# 
# This script will only consider detections with > 0.005 confidence at all times.
# The `threshold` you provide is only for rendering the results. If you need to
# see lower-confidence detections, you can change
# DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD.
# 
# Reference:
# https://github.com/tensorflow/models/blob/master/research/object_detection/inference/detection_inference.py
# 
########

#%% Constants, imports, environment

import argparse
import os
import statistics
import sys
import time
import warnings

import humanfriendly
from tqdm import tqdm

import md_utils.path_utils as path_utils
import md_visualization.visualization_utils as vis_utils

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings('ignore', '(Possibly )?corrupt EXIF data', UserWarning)

# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings('ignore', 'Metadata warning', UserWarning)

# Numpy FutureWarnings from tensorflow import
warnings.filterwarnings('ignore', category=FutureWarning)

# Useful hack to force CPU inference
# os.environ['CUDA_VISIBLE_DEVICES'] = '-1'


# An enumeration of failure reasons
FAILURE_INFER = 'Failure inference'
FAILURE_IMAGE_OPEN = 'Failure image access'

# Number of decimal places to round to for confidence and bbox coordinates
CONF_DIGITS = 3
COORD_DIGITS = 4

# Label mapping for MegaDetector
DEFAULT_DETECTOR_LABEL_MAP = {
    '1': 'animal',
    '2': 'person',
    '3': 'vehicle'  # available in megadetector v4+
}

# Should we allow classes that don't look anything like the MegaDetector classes?
#
# By default, we error if we see unfamiliar classes.
#
# TODO: the use of a global variable to manage this was fine when this was really
# experimental, but this is really sloppy now that we actually use this code for
# models other than MegaDetector.
USE_MODEL_NATIVE_CLASSES = False

# Each version of the detector is associated with some "typical" values
# that are included in output files, so that downstream applications can 
# use them as defaults.
DETECTOR_METADATA = {
    'v2.0.0':
        {'megadetector_version':'v2.0.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v3.0.0':
        {'megadetector_version':'v3.0.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v4.1.0':
        {'megadetector_version':'v4.1.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v5a.0.0':
        {'megadetector_version':'v5a.0.0',
         'typical_detection_threshold':0.2,
         'conservative_detection_threshold':0.05},
    'v5b.0.0':
        {'megadetector_version':'v5b.0.0',
         'typical_detection_threshold':0.2,
         'conservative_detection_threshold':0.05}
}

DEFAULT_RENDERING_CONFIDENCE_THRESHOLD = DETECTOR_METADATA['v5b.0.0']['typical_detection_threshold']
DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD = 0.005

DEFAULT_BOX_THICKNESS = 4
DEFAULT_BOX_EXPANSION = 0
DEFAULT_LABEL_FONT_SIZE = 16
DETECTION_FILENAME_INSERT = '_detections'


#%% Utility functions

def convert_to_tf_coords(array):
    """
    From [x1, y1, width, height] to [y1, x1, y2, x2], where x1 is x_min, x2 is x_max

    This is only used to keep the interface of the synchronous API.
    """
    
    x1 = array[0]
    y1 = array[1]
    width = array[2]
    height = array[3]
    x2 = x1 + width
    y2 = y1 + height
    return [y1, x1, y2, x2]


def get_detector_metadata_from_version_string(detector_version):
    """
    Given a MegaDetector version string (e.g. "v4.1.0"), return the metadata for
    the model.  Used for writing standard defaults to batch output files.
    """
    
    if detector_version not in DETECTOR_METADATA:
        print('Warning: no metadata for unknown detector version {}'.format(detector_version))
        default_detector_metadata = {
            'megadetector_version':'unknown'
        }
        return default_detector_metadata
    else:
        return DETECTOR_METADATA[detector_version]


def get_detector_version_from_filename(detector_filename):
    """
    Get the version number component of the detector from the model filename.  
    
    *detector_filename* will almost always end with one of the following:
        
    megadetector_v2.pb
    megadetector_v3.pb
    megadetector_v4.1 (not produed by run_detector_batch.py, only found in Azure Batch API output files)
    md_v4.1.0.pb
    md_v5a.0.0.pt
    md_v5b.0.0.pt
    
    ...for which we identify the version number as "v2.0.0", "v3.0.0", "v4.1.0", 
    "v4.1.0", "v5a.0.0", and "v5b.0.0", respectively.
    """
    
    fn = os.path.basename(detector_filename)
    known_model_versions = {'v2':'v2.0.0',
                            'v3':'v3.0.0',
                            'v4.1':'v4.1.0',
                            'v5a.0.0':'v5a.0.0',
                            'v5b.0.0':'v5b.0.0'}
    matches = []
    for s in known_model_versions.keys():
        if s in fn:
            matches.append(s)
    if len(matches) == 0:
        print('Warning: could not determine MegaDetector version for model file {}'.format(detector_filename))
        return 'unknown'
    elif len(matches) > 1:
        print('Warning: multiple MegaDetector versions for model file {}'.format(detector_filename))
        return 'multiple'
    else:
        return known_model_versions[matches[0]]
    

def get_typical_confidence_threshold_from_results(results):
    """
    Given the .json data loaded from a MD results file, determine a typical confidence
    threshold based on the detector version.
    """
    
    if 'detector_metadata' in results['info'] and \
        'typical_detection_threshold' in results['info']['detector_metadata']:
        default_threshold = results['info']['detector_metadata']['typical_detection_threshold']
    elif ('detector' not in results['info']) or (results['info']['detector'] is None):
        print('Warning: detector version not available in results file, using MDv5 defaults')
        detector_metadata = get_detector_metadata_from_version_string('v5a.0.0')
        default_threshold = detector_metadata['typical_detection_threshold']
    else:
        print('Warning: detector metadata not available in results file, inferring from MD version')
        detector_filename = results['info']['detector']
        detector_version = get_detector_version_from_filename(detector_filename)
        detector_metadata = get_detector_metadata_from_version_string(detector_version)
        default_threshold = detector_metadata['typical_detection_threshold']

    return default_threshold    

    
def is_gpu_available(model_file):
    """
    Decide whether a GPU is available, importing PyTorch or TF depending on the extension
    of model_file.  Does not actually load model_file, just uses that to determine how to check 
    for GPU availability.
    """
    
    if model_file.endswith('.pb'):
        import tensorflow.compat.v1 as tf
        gpu_available = tf.test.is_gpu_available()
        print('TensorFlow version:', tf.__version__)
        print('tf.test.is_gpu_available:', gpu_available)                
        return gpu_available
    elif model_file.endswith('.pt'):
        import torch
        gpu_available = torch.cuda.is_available()
        print('PyTorch reports {} available CUDA devices'.format(torch.cuda.device_count()))
        if not gpu_available:
            try:
                # mps backend only available in torch >= 1.12.0
                if torch.backends.mps.is_built and torch.backends.mps.is_available():
                    gpu_available = True
                    print('PyTorch reports Metal Performance Shaders are available')
            except AttributeError:
                pass
        return gpu_available
    else:
        raise ValueError('Unrecognized model file extension for model {}'.format(model_file))


def load_detector(model_file, force_cpu=False):
    """
    Load a TF or PT detector, depending on the extension of model_file.
    """
    
    start_time = time.time()
    if model_file.endswith('.pb'):
        from detection.tf_detector import TFDetector
        if force_cpu:
            raise ValueError('force_cpu is not currently supported for TF detectors, ' + \
                             'use CUDA_VISIBLE_DEVICES=-1 instead')
        detector = TFDetector(model_file)
    elif model_file.endswith('.pt'):
        from detection.pytorch_detector import PTDetector
        detector = PTDetector(model_file, force_cpu, USE_MODEL_NATIVE_CLASSES)        
    else:
        raise ValueError('Unrecognized model format: {}'.format(model_file))
    elapsed = time.time() - start_time
    print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))
    return detector


#%% Main function

def load_and_run_detector(model_file, image_file_names, output_dir,
                          render_confidence_threshold=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
                          crop_images=False, box_thickness=DEFAULT_BOX_THICKNESS, 
                          box_expansion=DEFAULT_BOX_EXPANSION, image_size=None,
                          label_font_size=DEFAULT_LABEL_FONT_SIZE
                          ):
    """
    Load and run detector on target images, and visualize the results.
    """
    
    if len(image_file_names) == 0:
        print('Warning: no files available')
        return

    print('GPU available: {}'.format(is_gpu_available(model_file)))
    
    detector = load_detector(model_file)

    detection_results = []
    time_load = []
    time_infer = []

    # Dictionary mapping output file names to a collision-avoidance count.
    #
    # Since we'll be writing a bunch of files to the same folder, we rename
    # as necessary to avoid collisions.
    output_filename_collision_counts = {}

    def input_file_to_detection_file(fn, crop_index=-1):
        """
        Creates unique file names for output files.

        This function does 3 things:
        1) If the --crop flag is used, then each input image may produce several output
            crops. For example, if foo.jpg has 3 detections, then this function should
            get called 3 times, with crop_index taking on 0, 1, then 2. Each time, this
            function appends crop_index to the filename, resulting in
                foo_crop00_detections.jpg
                foo_crop01_detections.jpg
                foo_crop02_detections.jpg

        2) If the --recursive flag is used, then the same file (base)name may appear
            multiple times. However, we output into a single flat folder. To avoid
            filename collisions, we prepend an integer prefix to duplicate filenames:
                foo_crop00_detections.jpg
                0000_foo_crop00_detections.jpg
                0001_foo_crop00_detections.jpg

        3) Prepends the output directory:
                out_dir/foo_crop00_detections.jpg

        Args:
            fn: str, filename
            crop_index: int, crop number

        Returns: output file path
        """
        
        fn = os.path.basename(fn).lower()
        name, ext = os.path.splitext(fn)
        if crop_index >= 0:
            name += '_crop{:0>2d}'.format(crop_index)
        fn = '{}{}{}'.format(name, DETECTION_FILENAME_INSERT, '.jpg')
        if fn in output_filename_collision_counts:
            n_collisions = output_filename_collision_counts[fn]
            fn = '{:0>4d}'.format(n_collisions) + '_' + fn
            output_filename_collision_counts[fn] += 1
        else:
            output_filename_collision_counts[fn] = 0
        fn = os.path.join(output_dir, fn)
        return fn

    # ...def input_file_to_detection_file()
    
    for im_file in tqdm(image_file_names):

        try:
            start_time = time.time()

            image = vis_utils.load_image(im_file)

            elapsed = time.time() - start_time
            time_load.append(elapsed)

        except Exception as e:
            print('Image {} cannot be loaded. Exception: {}'.format(im_file, e))
            result = {
                'file': im_file,
                'failure': FAILURE_IMAGE_OPEN
            }
            detection_results.append(result)
            continue

        try:
            start_time = time.time()

            result = detector.generate_detections_one_image(image, im_file,
                       detection_threshold=DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                       image_size=image_size)
            detection_results.append(result)

            elapsed = time.time() - start_time
            time_infer.append(elapsed)

        except Exception as e:
            print('An error occurred while running the detector on image {}. Exception: {}'.format(im_file, e))
            continue

        try:
            if crop_images:

                images_cropped = vis_utils.crop_image(result['detections'], image,
                                   confidence_threshold=render_confidence_threshold,
                                   expansion=box_expansion)

                for i_crop, cropped_image in enumerate(images_cropped):
                    output_full_path = input_file_to_detection_file(im_file, i_crop)
                    cropped_image.save(output_full_path)

            else:

                # Image is modified in place
                vis_utils.render_detection_bounding_boxes(result['detections'], image,
                            label_map=DEFAULT_DETECTOR_LABEL_MAP,
                            confidence_threshold=render_confidence_threshold,
                            thickness=box_thickness, expansion=box_expansion,
                            label_font_size=label_font_size)
                output_full_path = input_file_to_detection_file(im_file)
                image.save(output_full_path)

        except Exception as e:
            print('Visualizing results on the image {} failed. Exception: {}'.format(im_file, e))
            continue

    # ...for each image

    ave_time_load = statistics.mean(time_load)
    ave_time_infer = statistics.mean(time_infer)
    if len(time_load) > 1 and len(time_infer) > 1:
        std_dev_time_load = humanfriendly.format_timespan(statistics.stdev(time_load))
        std_dev_time_infer = humanfriendly.format_timespan(statistics.stdev(time_infer))
    else:
        std_dev_time_load = 'not available'
        std_dev_time_infer = 'not available'
    print('On average, for each image,')
    print('- loading took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_load),
                                                    std_dev_time_load))
    print('- inference took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_infer),
                                                      std_dev_time_infer))

# ...def load_and_run_detector()


#%% Command-line driver

def main():

    parser = argparse.ArgumentParser(
        description='Module to run an animal detection model on images')
    
    parser.add_argument(
        'detector_file',
        help='Path to TensorFlow (.pb) or PyTorch (.pt) detector model file')
    
    # Must specify either an image file or a directory
    group = parser.add_mutually_exclusive_group(required=True)    
    group.add_argument(
        '--image_file',
        help='Single file to process, mutually exclusive with --image_dir')
    group.add_argument(
        '--image_dir',
        help='Directory to search for images, with optional recursion by adding --recursive')
    
    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Recurse into directories, only meaningful if using --image_dir')
    
    parser.add_argument(
        '--output_dir',
        help='Directory for output images (defaults to same as input)')
    
    parser.add_argument(
        '--image_size',
        type=int,
        default=None,
        help=('Force image resizing to a (square) integer size (not recommended to change this)'))
    
    parser.add_argument(
        '--threshold',
        type=float,
        default=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
        help=('Confidence threshold between 0 and 1.0; only render' + 
              ' boxes above this confidence (defaults to {})'.format(
              DEFAULT_RENDERING_CONFIDENCE_THRESHOLD)))
    
    parser.add_argument(
        '--crop',
        default=False,
        action="store_true",
        help=('If set, produces separate output images for each crop, '
              'rather than adding bounding boxes to the original image'))
    
    parser.add_argument(
        '--box_thickness',
        type=int,
        default=DEFAULT_BOX_THICKNESS,
        help=('Line width (in pixels) for box rendering (defaults to {})'.format(
              DEFAULT_BOX_THICKNESS)))
    
    parser.add_argument(
        '--box_expansion',
        type=int,
        default=DEFAULT_BOX_EXPANSION,
        help=('Number of pixels to expand boxes by (defaults to {})'.format(
              DEFAULT_BOX_EXPANSION)))
    
    parser.add_argument(
        '--label_font_size',
        type=int,
        default=DEFAULT_LABEL_FONT_SIZE,
        help=('Label font size (defaults to {})'.format(
              DEFAULT_LABEL_FONT_SIZE)))
        
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    assert os.path.exists(args.detector_file), 'detector file {} does not exist'.format(
        args.detector_file)
    assert 0.0 < args.threshold <= 1.0, 'Confidence threshold needs to be between 0 and 1'

    if args.image_file:
        image_file_names = [args.image_file]
    else:
        image_file_names = path_utils.find_images(args.image_dir, args.recursive)

    print('Running detector on {} images...'.format(len(image_file_names)))

    if args.output_dir:
        os.makedirs(args.output_dir, exist_ok=True)
    else:
        if args.image_dir:
            args.output_dir = args.image_dir
        else:
            # but for a single image, args.image_dir is also None
            args.output_dir = os.path.dirname(args.image_file)

    load_and_run_detector(model_file=args.detector_file,
                          image_file_names=image_file_names,
                          output_dir=args.output_dir,
                          render_confidence_threshold=args.threshold,
                          box_thickness=args.box_thickness,
                          box_expansion=args.box_expansion,                          
                          crop_images=args.crop,
                          image_size=args.image_size,
                          label_font_size=args.label_font_size)


if __name__ == '__main__':
    main()


#%% Interactive driver

if False:

    #%%
    model_file = r'c:\temp\models\md_v4.1.0.pb'
    image_file_names = path_utils.find_images(r'c:\temp\demo_images\ssverymini')
    output_dir = r'c:\temp\demo_images\ssverymini'
    render_confidence_threshold = 0.8
    crop_images = True

    load_and_run_detector(model_file=model_file,
                          image_file_names=image_file_names,
                          output_dir=output_dir,
                          render_confidence_threshold=render_confidence_threshold,
                          crop_images=crop_images)



########
#
# visualize_detector_output.py
#
# Render images with bounding boxes annotated on them to a folder, based on a
# detector output result file (json). The original images can be local or in
# Azure Blob Storage.
#
########

#%% Imports

import argparse
import json
import os
import random
import sys
from typing import List, Optional

from tqdm import tqdm

from data_management.annotations.annotation_constants import (
    detector_bbox_category_id_to_name)  # here id is int
from md_visualization import visualization_utils as vis_utils
from md_utils.ct_utils import get_max_conf


#%% Constants

# convert category ID from int to str
DEFAULT_DETECTOR_LABEL_MAP = {
    str(k): v for k, v in detector_bbox_category_id_to_name.items()
}


#%% Main function

def visualize_detector_output(detector_output_path: str,
                              out_dir: str,
                              images_dir: str,
                              is_azure: bool = False,
                              confidence_threshold: float = 0.15,
                              sample: int = -1,
                              output_image_width: int = 700,
                              random_seed: Optional[int] = None,
                              render_detections_only: bool = False,
                              classification_confidence_threshold = 0.1,
                              html_output_file=None,
                              html_output_options=None) -> List[str]:
    
    """
    Draw bounding boxes on images given the output of the detector.

    Args:
        detector_output_path: str, path to detector output json file
        out_dir: str, path to directory for saving annotated images
        images_dir: str, path to local images dir, or a SAS URL to an Azure Blob
            Storage container
        is_azure: bool, whether images_dir points to an Azure URL
        confidence: float, threshold above which annotations will be rendered
        sample: int, maximum number of images to annotate, -1 for all
        output_image_width: int, width in pixels to resize images for display,
            set to -1 to use original image width
        random_seed: int, for deterministic image sampling when sample != -1
        render_detections_only: bool, only render images with above-threshold detections

    Returns: list of str, paths to annotated images
    """
    
    assert confidence_threshold >= 0 and confidence_threshold <= 1, (
        f'Confidence threshold {confidence_threshold} is invalid, must be in (0, 1).')

    assert os.path.exists(detector_output_path), (
        f'Detector output file does not exist at {detector_output_path}.')

    if is_azure:
        from md_utils import sas_blob_utils
    else:
        assert os.path.isdir(images_dir)

    os.makedirs(out_dir, exist_ok=True)


    ##%% Load detector output

    with open(detector_output_path) as f:
        detector_output = json.load(f)
    assert 'images' in detector_output, (
        'Detector output file should be a json with an "images" field.')
    images = detector_output['images']

    if 'detection_categories' in detector_output:
        print('Using custom label mapping')
        detector_label_map = detector_output['detection_categories']
    else:
        detector_label_map = DEFAULT_DETECTOR_LABEL_MAP        

    num_images = len(images)
    print(f'Detector output file contains {num_images} entries.')

    if sample > 0:
        assert num_images >= sample, (
            f'Sample size {sample} greater than number of entries '
            f'({num_images}) in detector result.')

        if random_seed is not None:
            images = sorted(images, key=lambda x: x['file'])
            random.seed(random_seed)

        random.shuffle(images)
        images = sorted(images[:sample], key=lambda x: x['file'])
        print(f'Sampled {len(images)} entries from the detector output file.')


    ##%% Load images, annotate them and save

    print('Rendering detections above a confidence threshold of {}'.format(
        confidence_threshold))
    
    num_saved = 0
    annotated_img_paths = []
    failed_images = []
    missing_images = []
    
    classification_label_map = None
    
    if 'classification_categories' in detector_output:
        classification_label_map = detector_output['classification_categories']
        
    for entry in tqdm(images):
        
        image_id = entry['file']

        if 'failure' in entry and entry['failure'] is not None:
            failed_images.append(image_id)
            continue

        assert 'detections' in entry and entry['detections'] is not None
        
        max_conf = get_max_conf(entry)
        if (max_conf < confidence_threshold) and render_detections_only:
            continue
        
        if is_azure:
            blob_uri = sas_blob_utils.build_blob_uri(
                container_uri=images_dir, blob_name=image_id)
            if not sas_blob_utils.check_blob_exists(blob_uri):
                container = sas_blob_utils.get_container_from_uri(images_dir)
                print(f'Image {image_id} not found in blob container '
                      f'{container}; skipped.')
                continue
            # BytesIO object
            image_obj, _ = sas_blob_utils.download_blob_to_stream(blob_uri)
        else:
            image_obj = os.path.join(images_dir, image_id)
            if not os.path.exists(image_obj):
                print(f'Image {image_id} not found in images_dir')
                missing_images.append(image_id)
                continue

        image = vis_utils.resize_image(
            vis_utils.open_image(image_obj), output_image_width)

        vis_utils.render_detection_bounding_boxes(
            entry['detections'], image, label_map=detector_label_map,
            classification_label_map = classification_label_map,
            confidence_threshold=confidence_threshold,
            classification_confidence_threshold=classification_confidence_threshold)

        for char in ['/', '\\', ':']:
            image_id = image_id.replace(char, '~')
        annotated_img_path = os.path.join(out_dir, f'anno_{image_id}')
        annotated_img_paths.append(annotated_img_path)
        image.save(annotated_img_path)
        num_saved += 1

        if is_azure:
            image_obj.close()

    # ...for each image
    
    print('Skipped {} failed images (of {})'.format(len(failed_images),len(images)))
    print('Skipped {} missing images (of {})'.format(len(missing_images),len(images)))
    
    print(f'Rendered detection results on {num_saved} images, '
          f'saved to {out_dir}')

    if html_output_file is not None:
        from md_utils import write_html_image_list
        write_html_image_list.write_html_image_list(html_output_file,annotated_img_paths,
                                                    options=html_output_options)
        
    return annotated_img_paths


#%% Command-line driver

def main() -> None:
    """Main function."""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Annotate the bounding boxes predicted by a detector above '
                    'some confidence threshold, and save the annotated images.')
    parser.add_argument(
        'detector_output_path', type=str,
        help='Path to json output file of the detector')
    parser.add_argument(
        'out_dir', type=str,
        help='Path to directory where the annotated images will be saved. '
             'The directory will be created if it does not exist.')
    parser.add_argument(
        '-c', '--confidence', type=float, default=0.15,
        help='Value between 0 and 1, indicating the confidence threshold '
             'above which to visualize bounding boxes')
    parser.add_argument(
        '-i', '--images_dir', type=str, default=None,
        help='Path to a local directory or a SAS URL (in double quotes) to an '
             'Azure blob storage container where images are stored. This '
             'serves as the root directory for image paths in '
             'detector_output_path. If an Azure URL, pass the -a/--is-azure '
             'flag. You can use Azure Storage Explorer to obtain a SAS URL.')
    parser.add_argument(
        '-a', '--is-azure', action='store_true',
        help='Flag that indidcates images_dir is an Azure blob storage '
             'container URL.')
    parser.add_argument(
        '-n', '--sample', type=int, default=-1,
        help='Number of images to be annotated and rendered. Set to -1 '
             '(default) to annotate all images in the detector output file. '
             'There may be fewer images if some are not found in images_dir.')
    parser.add_argument(
        '-w', '--output_image_width', type=int, default=700,
        help='Integer, desired width in pixels of the output annotated images. '
             'Use -1 to not resize. Default: 700.')
    parser.add_argument(
        '-r', '--random_seed', type=int, default=None,
        help='Integer, for deterministic order of image sampling')
    parser.add_argument(
        '-do', '--detections_only', action='store_true',
        help='Only render images with above-threshold detections (by default, '
             'both empty and non-empty images are rendered).')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    visualize_detector_output(
        detector_output_path=args.detector_output_path,
        out_dir=args.out_dir,
        confidence_threshold=args.confidence,
        images_dir=args.images_dir,
        is_azure=args.is_azure,
        sample=args.sample,
        output_image_width=args.output_image_width,
        random_seed=args.random_seed,
        render_detections_only=args.detections_only)


if __name__ == '__main__':
    main()



########
#
# md_to_coco_starter_code.py
#
# Partially-formed stub to get from MegaDetector output files to COCO Camera Traps data.
#
# Was actually written to convert *many* MD .json files to a single CCT file, hence 
# the loop over .json files.
#
# THIS CODE HAS NEVER BEEN RUN, it was added as a demonstration of how to do this.  YMMV.
#
# You may find a more polished, command-line-ready version of this code at:
#
# https://github.com/StewartWILDlab/mdtools
#
########

#%% Constants and imports

import os
import json
import uuid

from tqdm import tqdm
from PIL import Image
from collections import defaultdict

from md_visualization import visualize_db
from data_management.databases import sanity_check_json_db

json_file = 'some_file'
output_json_file = 'some_other_file'

# Images sizes are required to convert between absolute and relative coordinates,
# so we need to read the images.
image_base_folder = None

# Only required if you want to write a database preview
output_dir_base = '/some/folder'


#%% Create CCT dictionaries

images = []
annotations = []

image_ids_to_annotations = defaultdict(list)

# image_ids_to_images = {}

category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category
next_category_id = 1

print('Processing .json file {}'.format(json_file))

# Load .json annotations for this data set
with open(json_file, 'r') as f:
    data = f.read()        
data = json.loads(data)

categories_this_dataset = data['detection_categories']

# i_entry = 0; entry = data['images'][i_entry]
#
# PERF: Not exactly trivially parallelizable, but about 100% of the 
# time here is spent reading image sizes (which we need to do to get from 
# absolute to relative coordinates), so worth parallelizing.
for i_entry,entry in enumerate(tqdm(data['images'])):
    
    image_relative_path = entry['file']
    
    # Generate a unique ID from the path
    image_id = image_relative_path.split('.')[0].replace(
        '\\', '/').replace('/', '_').replace(' ', '_')
    
    im = {}
    im['id'] = image_id
    im['file_name'] = image_relative_path
    
    pil_image = Image.open(os.path.join(image_base_folder,image_relative_path))
    width, height = pil_image.size
    im['width'] = width
    im['height'] = height

    images.append(im)
    
    detections = entry['detections']
    
    # detection = detections[0]
    for detection in detections:
        
        category_name = categories_this_dataset[detection['category']]
        category_name = category_name.strip().lower()            
        category_name = category_name.replace(' ','_')        
        
        # Have we seen this category before?
        if category_name in category_name_to_category:
            category_id = category_name_to_category[category_name]['id']
        else:
            category_id = next_category_id
            category = {}
            category['id'] = category_id
            print('Adding category {}'.format(category_name))
            category['name'] = category_name
            category_name_to_category[category_name] = category
            next_category_id += 1
        
        # Create an annotation
        ann = {}        
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']    
        ann['category_id'] = category_id
        
        if category_id != 0:
            ann['bbox'] = detection['bbox']
            # MegaDetector: [x,y,width,height] (normalized, origin upper-left)
            # CCT: [x,y,width,height] (absolute, origin upper-left)
            ann['bbox'][0] = ann['bbox'][0] * im['width']
            ann['bbox'][1] = ann['bbox'][1] * im['height']
            ann['bbox'][2] = ann['bbox'][2] * im['width']
            ann['bbox'][3] = ann['bbox'][3] * im['height']
        else:
            assert(detection['bbox'] == [0,0,0,0])
        annotations.append(ann)
        image_ids_to_annotations[im['id']].append(ann)
        
    # ...for each detection

# ...for each image
            
print('Finished creating CCT dictionaries')

# Remove non-reviewed images and associated annotations


#%% Create info struct

info = dict()
info['year'] = 2020
info['version'] = 1.0
info['description'] = 'Fun With Camera Traps'
info['contributor'] = 'Somebody'


#%% Write .json output

categories = list(category_name_to_category.values())

json_data = {}

json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=2)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
    len(images), len(annotations), len(categories)))


#%% Clean start

### Everything after this should work from a clean start ###


#%% Validate output

fn = output_json_file
options = sanity_check_json_db.SanityCheckOptions()
options.baseDir = image_base_folder
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False
sorted_categories, data, errors = sanity_check_json_db.sanity_check_json_db(fn, options)


#%% Preview animal labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 3000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file, image_db = visualize_db.process_images(db_path=output_json_file,
                                                         output_dir=os.path.join(
                                                         output_dir_base, 'preview_animals'),
                                                         image_base_dir=image_base_folder,
                                                         options=viz_options)
os.startfile(html_output_file)


#%% Preview empty labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 3000
viz_options.trim_to_images_with_bboxes = False
# viz_options.classes_to_exclude = ['empty','human']
viz_options.classes_to_include = ['empty']
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file, image_db = visualize_db.process_images(db_path=output_json_file,
                                                         output_dir=os.path.join(
                                                         output_dir_base, 'preview_empty'),
                                                         image_base_dir=image_base_folder,
                                                         options=viz_options)
os.startfile(html_output_file)



########
#
# csv_to_json.py
#
# Converts CSV to JSON format for label specification.
# 
# There are 3 possible values for the 'type' column in the CSV:
#
# - "row": this selects a specific rowfrom the master taxonomy CSV
#     content syntax: <dataset_name>|<dataset_label>
#
# - "datasettaxon": this selects all animals in a taxon from a particular dataset
#     content syntax: <dataset_name>|<taxon_level>|<taxon_name>
#
# - <taxon_level>: this selects all animals in a taxon across all datasets
#     content syntax: <taxon_name>
# 
# Example CSV input:
#
"""   
    # comment lines starting with '#' are allowed
    output_label,type,content

    cervid,row,idfg|deer
    cervid,row,idfg|elk
    cervid,row,idfg|prong
    cervid,row,idfg_swwlf_2019|elk
    cervid,row,idfg_swwlf_2019|muledeer
    cervid,row,idfg_swwlf_2019|whitetaileddeer
    cervid,max_count,50000

    cervid,family,cervidae
    cervid,datasettaxon,idfg|family|cervidae
    cervid,datasettaxon,idfg_swwlf_2019|family|cervidae

    bird,row,idfg_swwlf_2019|bird
    bird,class,aves
    bird,max_count,50000
    bird,prioritize,"[['idfg_swwlf_2019'], ['idfg']]"

    !bird,row,idfg_swwlf_2019|turkey
    !bird,genus,meleagris
"""
#
# Example JSON output:
#
"""    
    {
        "cervid": {
            "dataset_labels": {
                "idfg": ["deer", "elk", "prong"],
                "idfg_swwlf_2019": ["elk", "muledeer", "whitetaileddeer"]
            },
            "taxa": [
                {
                    "level": "family",
                    "name": "cervidae"
                },
                {
                    "level": "family",
                    "name": "cervidae"
                    "datasets": ["idfg"]
                },
                {
                    "level": "family",
                    "name": "cervidae"
                    "datasets": ["idfg_swwlf_2019"]
                }
            ],
            "max_count": 50000
        },
        "bird": {
            "dataset_labels": {
                "idfg_swwlf_2019": ["bird"]
            },
            "taxa": [
                {
                    "level": "class",
                    "name": "aves"
                }
            ],
            "exclude": {
                "dataset_labels": {
                    "idfg_swwlf_2019": ["turkey"]
                },
                "taxa": [
                    {
                        "level": "genus",
                        "name": "meleagris"
                    }
                ]
            },
            "max_count": "50000",
            "prioritize": [
                ["idfg_swwlf_2019"],
                ["idfg"]
            ],
        }
    }
"""
#
########

#%% Imports

from __future__ import annotations

import argparse
from collections import defaultdict
import json
from typing import Any

import pandas as pd


#%% Main function

def main():
    args = _parse_args()
    js = csv_to_jsondict(args.input_csv_file)
    for label in js:
        js[label] = order_spec_dict(js[label])
    with open(args.output_json_path, 'w') as f:
        json.dump(js, f, indent=args.json_indent)


#%% Support functions
def parse_csv_row(obj: dict[str, Any], rowtype: str, content: str) -> None:
    """
    Parses a row in the CSV.
    """
    
    if rowtype == 'row':
        if 'dataset_labels' not in obj:
            obj['dataset_labels'] = defaultdict(list)

        assert '|' in content
        ds, ds_label = content.split('|')
        obj['dataset_labels'][ds].append(ds_label)

    elif rowtype == 'datasettaxon':
        if 'taxa' not in obj:
            obj['taxa'] = []

        assert '|' in content
        ds, taxon_level, taxon_name = content.split('|')
        obj['taxa'].append({
            'level': taxon_level,
            'name': taxon_name,
            'datasets': [ds]
        })

    elif rowtype == 'max_count':
        obj['max_count'] = int(content)

    elif rowtype == 'prioritize':
        obj['prioritize'] = eval(content)  # pylint: disable=eval-used

    else:
        if 'taxa' not in obj:
            obj['taxa'] = []

        taxon_level = rowtype
        taxon_name = content
        obj['taxa'].append({
            'level': taxon_level,
            'name': taxon_name
        })


def csv_to_jsondict(csv_path: str) -> dict[str, dict[str, Any]]:
    """
    Converts CSV to json-style dictionary.
    """
    
    df = pd.read_csv(csv_path, comment='#', skip_blank_lines=True)
    assert (df.columns == ['output_label', 'type', 'content']).all()

    js: defaultdict[str, dict[str, Any]] = defaultdict(dict)

    for i in df.index:
        label, rowtype, content = df.loc[i]
        if label.startswith('!'):
            label = label[1:]
            if 'exclude' not in js[label]:
                js[label]['exclude'] = {}
            obj = js[label]['exclude']
        else:
            obj = js[label]
        parse_csv_row(obj, rowtype, content)

    return js


def order_spec_dict(spec_dict: dict[str, Any]) -> dict[str, Any]:
    """
    Returns spec_dict with keys in a specific order.
    """
    
    if 'exclude' in spec_dict:
        spec_dict['exclude'] = order_spec_dict(spec_dict['exclude'])
    ordered_spec_dict: dict[str, Any] = {}
    for key in ['dataset_labels', 'taxa', 'exclude', 'max_count', 'prioritize']:
        if key in spec_dict:
            ordered_spec_dict[key] = spec_dict[key]
    return ordered_spec_dict


#%% Command-line driver

def _parse_args() -> argparse.Namespace:
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Converts CSV to JSON format for label specification.')
    parser.add_argument(
        'input_csv_file',
        help='path to CSV file containing label specification')
    parser.add_argument(
        'output_json_path',
        help='path to save converted JSON file')
    parser.add_argument(
        '--json-indent', type=int,
        help='number of spaces to use for JSON indent (default no indent)')
    return parser.parse_args()


if __name__ == '__main__':
    main()



########
#
# remove_corrupted_images_from_db.py
#
# Given a coco-camera-traps .json file, checks all images for corruption
# and generates a new .json file that only contains the non-corrupted images.
#
########

#%% Imports and constants

import argparse
import gc
import json
import os
import time

from multiprocessing.pool import ThreadPool

import humanfriendly
import numpy as np
import tensorflow as tf

N_THREADS = 16
DEBUG_MAX_IMAGES = -1
IMAGE_PRINT_FREQUENCY = 10


#%% Function definitions

def check_images(images, image_file_root):    
    """
    Checks all the images in [images] for corruption using TF.
    
    [images] is a list of image dictionaries, as they would appear in COCO
    files.
    
    Returns a dictionary mapping image IDs to booleans. 
    """    
    
    # I sometimes pass in a list of images, sometimes a dict with a single
    # element mapping a job ID to the list of images
    if isinstance(images,dict):
        assert(len(images) == 1)
        jobID = list(images.keys())[0]
        images = images[jobID]
    else:
        jobID = 0
        
    keep_im = {im['id']:True for im in images}
            
    count = 0
    nImages = len(images)
    
    # We're about to start a lot of TF sessions, and we don't want gobs 
    # of debugging information printing out for every session.
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    config=tf.ConfigProto(log_device_placement=False)
    
    # At some point we were creating a single session and looping over images
    # within that session, but the only way I found to reliably not run out
    # of GPU memory was to create a session per image and gc.collect() after
    # each session.
    for iImage,im in enumerate(images):

        with tf.Session(config=config) as sess:

            if ((DEBUG_MAX_IMAGES > 0) and (iImage >= DEBUG_MAX_IMAGES)):
                print('Breaking after {} images'.format(DEBUG_MAX_IMAGES))
                break
            
            if (count % IMAGE_PRINT_FREQUENCY == 0):
                print('Job {}: processed {} of {} images'.format(jobID,count,nImages))
                
            count += 1
            image_file = os.path.join(image_file_root,im['file_name'])
            assert(os.path.isfile(image_file))
            
            try:
                image_data = tf.gfile.FastGFile(image_file,'rb').read()
                image = tf.image.decode_jpeg(image_data)
                sess.run(image)
            except:
                keep_im[im['id']] = False

        gc.collect()
        
    return keep_im


def remove_corrupted_images_from_database(data, image_file_root):
    """
    Given the COCO database [data], checks all images for corruption using
    TF, and returns a subset of [data] containing only non-corrupted images.
    """
    
    # Map Image IDs to boolean (should I keep this image?)
    images = data['images']
    
    if (N_THREADS == 1):
        
        keep_im = check_images(images,image_file_root)
        
    else:
        
        start = time.time()
        imageChunks = np.array_split(images,N_THREADS)
        # Convert to lists, append job numbers to the image lists
        for iChunk in range(0,len(imageChunks)):
            imageChunks[iChunk] = list(imageChunks[iChunk])
            imageChunks[iChunk] = {iChunk:imageChunks[iChunk]}
        pool = ThreadPool(N_THREADS)
        # results = pool.imap_unordered(lambda x: fetch_url(x,nImages), indexedUrlList)
        results = pool.map(lambda x: check_images(x,image_file_root), imageChunks)
        processingTime = time.time() - start
        
        # Merge results
        keep_im = {}
        for d in results:
            keep_im.update(d)
        bValid = keep_im.values()
            
        print("Checked image corruption in {}, found {} invalid images (of {})".format(
                humanfriendly.format_timespan(processingTime),
                len(bValid)-sum(bValid),len(bValid)))
        
    data['images'] = [im for im in data['images'] if keep_im[im['id']]]
    data['annotations'] = [ann for ann in data['annotations'] if keep_im[ann['image_id']]]
    
    return data


#%% Interactive driver

if False:    
    
    #%%
    
    # base_dir = r'D:\temp\snapshot_serengeti_tfrecord_generation'
    base_dir = r'/data/ss_corruption_check'
    input_file = os.path.join(base_dir,'imerit_batch7_renamed.json')
    output_file = os.path.join(base_dir,'imerit_batch7_renamed_uncorrupted.json')
    image_file_root = os.path.join(base_dir,'imerit_batch7_images_renamed')
    assert(os.path.isfile(input_file))
    assert(os.path.isdir(image_file_root))

    # Load annotations
    with open(input_file,'r') as f:
            data = json.load(f)    
            
    # Check for corruption
    data_uncorrupted = remove_corrupted_images_from_database(data,image_file_root)
    
    # Write out only the uncorrupted data
    json.dump(data_uncorrupted, open(output_file,'w'))
    
    
#%% Command-line driver

def parse_args():
    
    parser = argparse.ArgumentParser(description = 'Remove images from a .json file that can''t be opened in TF')

    parser.add_argument('--input_file', dest='input_file',
                         help='Path to .json database that includes corrupted jpegs',
                         type=str, required=True)
    parser.add_argument('--image_file_root', dest='image_file_root',
                         help='Path to image files',
                         type=str, required=True)
    parser.add_argument('--output_file', dest='output_file',
                         help='Path to store uncorrupted .json database',
                         type=str, required=True)

    args = parser.parse_args()
    return args
 

def main():
    
    args = parse_args()
    print('Reading input file')
    with open(args.input_file,'r') as f:
        data = json.load(f)
    print('Removing corrupted images from database')
    uncorrupted_data = remove_corrupted_images_from_database(data, args.image_file_root)

    json.dump(uncorrupted_data, open(args.output_file,'w'))


if __name__ == '__main__':    
    main()



########
#
# missouri_to_json.py
#
# Create .json files from the original source files for the Missouri Camera Traps
# data set.  Metadata was provided here in two formats:
#
# 1) In one subset of the data, folder names indicated species names.  In Set 1,
#    there are no empty sequences.  Set 1 has a metadata file to indicate image-level
#    bounding boxes.
#
# 2) A subset of the data (overlapping with (1)) was annotated with bounding
#    boxes, specified in a whitespace-delimited text file.  In set 2, there are
#    some sequences omitted from the metadata file, which implied emptiness.
# 
# In the end, set 2 labels were not reliable enough to publish, so LILA includes only set 1.
#
########

#%% Constants and imports

import json
import os
import uuid
import time
import humanfriendly
import warnings
import ntpath
import datetime
from PIL import Image

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings("ignore", "(Possibly )?corrupt EXIF data", UserWarning)
# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings("ignore", "Metadata warning", UserWarning)

# Filenames will be stored in the output .json relative to this base dir
baseDir = r'g:\temp\mct'
outputJsonFilenameSet1 = os.path.join(baseDir,'missouri_camera_traps_set1.json')
outputJsonFilenameSet2 = os.path.join(baseDir,'missouri_camera_traps_set2.json')
outputEncoding = 'utf-8'
fileListJsonFilename = os.path.join(baseDir,'images.json')

# This will be a list of filenames that need re-annotation due to redundant boxes
set1RedundantBoxListFilename = os.path.join(baseDir,'mct_images_with_redundant_boxes.txt')

set1BaseDir = os.path.join(baseDir,'Set1')
set2BaseDir = os.path.join(baseDir,'Set2')

metadataFilenameSet1 = os.path.join(set1BaseDir,'labels.txt')
metadataFilenameSet2 = os.path.join(set2BaseDir,'labels.txt')

assert(os.path.isdir(baseDir))
assert(os.path.isfile(metadataFilenameSet1))

if os.path.isfile(metadataFilenameSet2):
    process_set2 = True
else:
    print('Set 2 metadata not available, processing only set 1')
    process_set2 = False


info = {}
info['year'] = 2019
info['version'] = '1.2'
info['description'] = 'Missouri Camera Traps (set 1)'
info['contributor'] = ''
info['date_created'] = str(datetime.date.today())
infoSet1 = info

if process_set2:
    info = {}
    info['year'] = 2019
    info['version'] = '1.1'
    info['description'] = 'Missouri Camera Traps (set 2)'
    info['contributor'] = ''
    info['date_created'] = str(datetime.date.today())
    infoSet2 = info

maxFiles = -1
emptyCategoryId = 0
emptyCategoryName = 'empty'


#%% Enumerate files, read image sizes (both sets)

# Takes a few minutes, since we're reading image sizes.

# Each element will be a list of relative path/full path/width/height
fileInfo = []
nonImages = []
nFiles = 0

relPathToIm = {}
imageIdToImage = {}

set1ImageIDs = []
set2ImageIDs = []
   
sequenceIDtoCount = {}

print('Enumerating files from {} to {}'.format(baseDir,fileListJsonFilename))

startTime = time.time()

for root, subdirs, files in os.walk(baseDir):
            
    if root == baseDir:
        continue
    
    bn = ntpath.basename(root)
    
    # Only process leaf nodes corresponding to sequences, which look like:
    #
    # E:\wildlife_data\missouri_camera_traps\Set1\1.02-Agouti\SEQ75583
    # E:\wildlife_data\missouri_camera_traps\Set2\p1d101
    #
    if ('Set1' in root and 'SEQ' in bn) or ('Set2' in root and bn.startswith('p')):
        sequenceID = bn
        assert sequenceID not in sequenceIDtoCount
        sequenceIDtoCount[sequenceID] = 0
    else:
        print('Skipping folder {}:{}'.format(root,bn))
        continue
        # assert len(files) <= 2
    
    for fname in files:
  
        nFiles = nFiles + 1
        if maxFiles >= 0 and nFiles > maxFiles:            
            print('Warning: early break at {} files'.format(maxFiles))
            break
        
        fullPath = os.path.join(root,fname)            
        relativePath = os.path.relpath(fullPath,baseDir)
        
        if maxFiles >= 0:
            print(relativePath)
    
        h = -1
        w = -1

        # Read the image
        try:
        
            im = Image.open(fullPath)
            h = im.height
            w = im.width
            
        except:
            
            # Not an image...
            continue
        
        # Store file info
        im = {}
        im['id'] = str(uuid.uuid1())
        im['file_name'] = relativePath
        im['height'] = h
        im['width'] = w
        im['location'] = 'missouri_camera_traps'
        
        im['seq_id'] = sequenceID
        im['seq_num_frames'] = -1
        
        frame_number = sequenceIDtoCount[sequenceID]
        im['frame_num'] = frame_number
        sequenceIDtoCount[sequenceID] = sequenceIDtoCount[sequenceID] + 1
        
        imageIdToImage[im['id']] = im
        relPathToIm[relativePath] = im
        
        if 'Set1' in relativePath:
            set1ImageIDs.append(im['id'])
        elif 'Set2' in relativePath:
            set2ImageIDs.append(im['id'])
        else:
            raise Exception('Oops, can''t assign this image to a set')
            
    # ...if we didn't hit the max file limit, keep going
    
    else:
        
        continue
    
    break

# ...for each file

elapsed = time.time() - startTime
print('Finished file enumeration in {}'.format(
      humanfriendly.format_timespan(elapsed)))


#%% Add sequence lengths (both sets)
    
for imageID in imageIdToImage:
    
    im = imageIdToImage[imageID]
    sequenceID = im['seq_id']
    seq_num_frames = sequenceIDtoCount[sequenceID]
    assert(im['seq_num_frames'] == -1)
    im['seq_num_frames'] = seq_num_frames
    

#%% Load the set 1 metadata file

with open(metadataFilenameSet1) as f:
    metadataSet1Lines = f.readlines()

correctedFiles = []
missingFilesSet1 = []

metadataSet1Lines = [x.strip() for x in metadataSet1Lines] 

# List of lists, length varies according to number of bounding boxes
#
# Preserves original ordering
metadataSet1 = []

relPathToMetadataSet1 = {}

# iLine = 0; line = metadataSet1Lines[0]
for iLine,line in enumerate(metadataSet1Lines):
    
    tokens = line.split()
    nTokens = len(tokens)
    
    # Lines should be filename, number of bounding boxes, boxes (four values per box)
    assert ((nTokens - 2) % 4) == 0
    relPath = tokens[0].replace('/',os.sep).replace('\\',os.sep)
    relPath = os.path.join('Set1',relPath)
    absPath = os.path.join(baseDir,relPath)
    
    originalAbsPath = absPath
    originalRelPath = relPath
        
    if not os.path.isfile(absPath):
        
        absPath = absPath.replace('IMG','IMG_')     
        relPath = relPath.replace('IMG','IMG_')        
        correctedFiles.append([relPath,originalRelPath,absPath,originalAbsPath])
        
    if not os.path.isfile(absPath):
        
        missingFilesSet1.append([originalRelPath,originalAbsPath])
        
    else:
        
        metadataSet1.append(tokens)
        relPathToMetadataSet1[relPath] = tokens
        
        # Make sure we have image info for this image
        assert relPath in relPathToIm

print('Corrected {} paths, missing {} images of {}'.format(len(correctedFiles),
      len(missingFilesSet1),len(metadataSet1Lines)))


#%% Print missing files from Set 1 metadata

# Manual changes I made to the metadata file:
#
# 'IMG' --> 'IMG_'
# Red_Brocket_Deer --> Red_Deer
# European-Hare --> European_Hare
# Wood-Mouse --> Wood_Mouse
# Coiban-Agouti --> Coiban_Agouti

print('Missing files in Set 1:\n')
for iFile,fInfo in enumerate(missingFilesSet1):
    print(fInfo[0])
    

#%% Load the set 2 metadata file

if process_set2:
    
    # This metadata file contains most (but not all) images, and a class label (person/animal/empty)
    # for each, plus bounding boxes.
    
    with open(metadataFilenameSet2) as f:
        metadataSet2Lines = f.readlines()
    
    metadataSet2Lines = [x.strip() for x in metadataSet2Lines] 
    
    set2ClassMappings = {0:'human',1:'animal'}
    
    # List of lists, length varies according to number of bounding boxes
    #
    # Preserves original ordering
    metadataSet2 = []
    
    relPathToMetadataSet2 = {}
    
    # Create class IDs for each *sequence*, which we'll use to attach classes to 
    # images for which we don't have metadata
    #
    # This only contains mappings for sequences that appear in the metadata.
    set2SequenceToClass = {}
    
    missingFilesSet2 = []
    
    # iLine = 0; line = metadataSet2Lines[0]
    for iLine,line in enumerate(metadataSet2Lines):
        
        tokens = line.split()
        nTokens = len(tokens)
        
        # Lines should be filename, number of bounding boxes, labeled boxes (five values per box)
        #
        # Empty images look like filename\t0\t0
        assert (nTokens == 3 and tokens[1] == '0' and tokens[2] == '0') or (((nTokens - 2) % 5) == 0)
        relPath = tokens[0].replace('/',os.sep).replace('\\',os.sep)
        relPath = os.path.join('Set2',relPath)
        absPath = os.path.join(baseDir,relPath)
    
        # E.g. 'Set2\\p1d101\\p1d101s100i10.JPG'
        fnTokens = os.path.normpath(relPath).split(os.sep)
        sequenceID = fnTokens[-2]
        
        # Make sure we don't have mixed classes within an image
        nBoxes = int(tokens[1])
        imageClass = None
        if nBoxes == 0:
            imageClass = 'empty'
        else:
            for iBox in range(0,nBoxes):
                boxClass = int(tokens[2+5*(iBox)])
                boxClass = set2ClassMappings[boxClass]
                if imageClass is None:
                    imageClass = boxClass
                elif boxClass != imageClass:
                    imageClass = 'mixed'                
        
        assert imageClass is not None
        
        # Figure out what class this *sequence* is, so we know how to handle unlabeled
        # images from this sequence
        if sequenceID in set2SequenceToClass:
            
            sequenceClass = set2SequenceToClass[sequenceID]
            
            # Can't un-do a mixed sequence
            if sequenceClass == 'mixed':
                pass
            
            # Previously-empty sequences get the image class label
            elif sequenceClass == 'empty':
                if imageClass != 'empty':
                    sequenceClass = imageClass
            
            # If the sequence has a non-empty class, possibly change it
            else:
                if imageClass == 'empty':
                    pass
                elif imageClass != sequenceClass:
                    sequenceClass = imageClass
            
            set2SequenceToClass[sequenceID] = sequenceClass
            
        else:
            
            set2SequenceToClass[sequenceID] = imageClass
        
        if not os.path.isfile(absPath):
            missingFilesSet2.append(absPath)
            continue
        
        metadataSet2.append(tokens)
        relPathToMetadataSet2[relPath] = tokens
            
        # Make sure we have image info for this image
        assert relPath in relPathToIm
    
    # ...for each line in the set 2 metadata file
        
    print('Missing {} of {} files in set 2'.format(len(missingFilesSet2),len(metadataSet2Lines)))
    
    if False:
        for iFile,filename in enumerate(missingFilesSet2):
            print(filename)
    

#%% What Set 2 images do I not have metadata for?
    
if process_set2:
    
    # These are *mostly* empty images
    
    if True:
    
        set2UnlabeledImageIndices = []
        set2ImageIDsUnlabeled = []
        
        # iImage = 0; imageID = set2ImageIDs[iImage]
        for iImage,imageID in enumerate(set2ImageIDs):
            
            im = imageIdToImage[imageID]
            if not im['file_name'] in relPathToMetadataSet2:
                set2UnlabeledImageIndices.append(iImage)
                set2ImageIDsUnlabeled.append(imageID)
                
        print('{} of {} files in set 2 lack labels'.format(len(set2UnlabeledImageIndices),len(set2ImageIDs)))
        
        set2ImageIDsLabeled = [x for i,x in enumerate(set2ImageIDs) if i not in set2UnlabeledImageIndices]    
        
        assert len(set2ImageIDsLabeled) + len(set2ImageIDsUnlabeled) == len(set2ImageIDs)
    

#%% Create categories and annotations for set 1

imagesSet1 = []
categoriesSet1 = []
annotationsSet1 = []

categoryNameToId = {}
idToCategory = {}

# Though we have no empty sequences, we do have empty images in this set
emptyCat = {}
emptyCat['id'] = emptyCategoryId
emptyCat['name'] = emptyCategoryName
emptyCat['count'] = 0
categoriesSet1.append(emptyCat) 

nextCategoryId = emptyCategoryId + 1
    
nFoundMetadata = 0
nTotalBoxes = 0
nImageLevelEmpties = 0
nSequenceLevelAnnotations = 0
nRedundantBoxes = 0

imageIDsWithRedundantBoxes = set()

# For each image
#
# iImage = 0; imageID = set1ImageIDs[iImage]
for iImage,imageID in enumerate(set1ImageIDs):
    
    im = imageIdToImage[imageID]
    imagesSet1.append(im)
    
    # E.g. Set1\\1.80-Coiban_Agouti\\SEQ83155\\SEQ83155_IMG_0010.JPG
    relPath = im['file_name']

    # Find the species name
    tokens = os.path.normpath(relPath).split(os.sep)
    speciesTag = tokens[1]
    tokens = speciesTag.split('-',1)
    assert(len(tokens) == 2)
    categoryName = tokens[1].lower()
    
    category = None
    categoryId = None
    
    if categoryName not in categoryNameToId:
        
        categoryId = nextCategoryId
        nextCategoryId += 1
        categoryNameToId[categoryName] = categoryId
        newCat = {}
        newCat['id'] = categoryNameToId[categoryName]
        newCat['name'] = categoryName
        newCat['count'] = 0
        categoriesSet1.append(newCat) 
        idToCategory[categoryId] = newCat
        category = newCat
        
    else:
        
        categoryId = categoryNameToId[categoryName]
        category = idToCategory[categoryId]
        
        # This image may still be empty...
        # category['count'] = category['count'] + 1
                
    # If we have bounding boxes, create image-level annotations    
    if relPath in relPathToMetadataSet1:
        
        nFoundMetadata += 1
        
        # filename, number of bounding boxes, boxes (four values per box)
        imageMetadata = relPathToMetadataSet1[relPath]
        
        # Make sure the relative filename matches, allowing for the fact that
        # some of the filenames in the metadata aren't quite right
        fn = imageMetadata[0]
        s = relPath.replace('Set1','').replace('\\','/')[1:]
        if (fn != s):
            s = s.replace('IMG_','IMG')
            
        assert(fn == s)
        
        nBoxes = int(imageMetadata[1])
        im['n_boxes'] = nBoxes
        
        if nBoxes == 0:
            
            ann = {}
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = im['id']
            ann['category_id'] = emptyCategoryId
            ann['sequence_level_annotation'] = False
            annotationsSet1.append(ann)
            emptyCat['count'] = emptyCat['count'] + 1
            nImageLevelEmpties += 1
            
        else:
            
            # This image is non-empty
            category['count'] = category['count'] + 1                        
            
            for iBox in range(0,nBoxes):
                                
                boxCoords = imageMetadata[2+(iBox*4):6+(iBox*4)]
                boxCoords = list(map(int, boxCoords))
                
                # Some redundant bounding boxes crept in, don't add them twice
                bRedundantBox = False
                
                # Check this bbox against previous bboxes
                #
                # Inefficient?  Yes.  In an important way?  No.
                for iBoxComparison in range(0,iBox):
                    assert iBox != iBoxComparison                        
                    boxCoordsComparison = imageMetadata[2+(iBoxComparison*4):6+(iBoxComparison*4)]
                    boxCoordsComparison = list(map(int, boxCoordsComparison))
                    if boxCoordsComparison == boxCoords:
                        print('Warning: redundant box on image {}'.format(fn))
                        bRedundantBox = True
                        nRedundantBoxes += 1
                        break
                
                if bRedundantBox:
                    imageIDsWithRedundantBoxes.add(im['id'])
                    continue
                    
                # Bounding box values are in absolute coordinates, with the origin 
                # at the upper-left of the image, as [xmin1 ymin1 xmax1 ymax1].
                #
                # Convert to floats and to x/y/w/h, as per CCT standard
                bboxW = boxCoords[2] - boxCoords[0]
                bboxH = boxCoords[3] - boxCoords[1]
                
                box = [boxCoords[0], boxCoords[1], bboxW, bboxH]
                box = list(map(float, box))
                
                ann = {}
                ann['id'] = str(uuid.uuid1())
                ann['image_id'] = im['id']
                ann['category_id'] = categoryId
                ann['sequence_level_annotation'] = False
                ann['bbox'] = box
                annotationsSet1.append(ann)
                nTotalBoxes += 1
            
            # ...for each box
            
        # if we do/don't have boxes for this image
        
    # Else create a sequence-level annotation
    else:
        
        ann = {}
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']
        ann['category_id'] = categoryId
        ann['sequence_level_annotation'] = True
        annotationsSet1.append(ann)
        nSequenceLevelAnnotations += 1
        
# ...for each image
        
print('Finished processing set 1, found metadata for {} of {} images\nCreated {} annotations and {} boxes in {} categories\nFound {} redundant annotations'.format(
        nFoundMetadata,len(set1ImageIDs),len(annotationsSet1),nTotalBoxes,len(categoriesSet1),nRedundantBoxes))

assert len(annotationsSet1) == nSequenceLevelAnnotations + nTotalBoxes + nImageLevelEmpties
assert len(set1ImageIDs) == nSequenceLevelAnnotations + nFoundMetadata

print('Found {} images with redundant boxes'.format(len(imageIDsWithRedundantBoxes)))


#%% Write out the list of images with redundant boxes

imageFileNamesWithRedundantBoxes = []
for image_id in imageIDsWithRedundantBoxes:
    im = imageIdToImage[image_id]
    imageFileNamesWithRedundantBoxes.append(im['file_name'])
imageFileNamesWithRedundantBoxes.sort()

with open(set1RedundantBoxListFilename,'w') as f:
    for fn in imageFileNamesWithRedundantBoxes:
        f.write(fn + '\n')
    
    
#%% Create categories and annotations for set 2

if process_set2:

    imagesSet2 = []
    categoriesSet2 = []
    annotationsSet2 = []
    
    categoryNameToId = {}
    idToCategory = {}
    
    emptyCat = {}
    assert emptyCategoryId == 0
    emptyCat['id'] = emptyCategoryId
    emptyCat['name'] = emptyCategoryName
    emptyCat['count'] = 0
    categoriesSet2.append(emptyCat) 
    
    humanCat = {}
    humanCat['id'] = 1
    humanCat['name'] = 'human'
    humanCat['count'] = 0
    categoriesSet2.append(humanCat) 
    
    animalCat = {}
    animalCat['id'] = 2
    animalCat['name'] = 'animal'
    animalCat['count'] = 0
    categoriesSet2.append(animalCat) 
    
    unknownCat = {}
    unknownCat['id'] = 3
    unknownCat['name'] = 'unknown'
    unknownCat['count'] = 0
    categoriesSet2.append(unknownCat) 
    
    categoryMappingsSet2 = {0:humanCat,1:animalCat}
    sequenceClassCategoryMappingsSet2 = {'human':humanCat,'animal':animalCat,
                                         'mixed':unknownCat,'empty':emptyCat}
    
    nFoundMetadata = 0
    nImageLevelEmpties = 0
    nSequenceLevelAnnotations = 0
    nTotalBoxes = 0
    
    # For each image
    #
    # iImage = 0; imageID = set2ImageIDs[iImage]
    for iImage,imageID in enumerate(set2ImageIDs):
        
        im = imageIdToImage[imageID]    
        imagesSet2.append(im)
        
        # E.g. 'Set2\\p1d100\\p1d100s10i1.JPG'
        relPath = im['file_name']
    
        # Find the sequence ID, sanity check filename against what we stored
        tokens = os.path.normpath(relPath).split(os.sep)
        sequenceID = tokens[1]
        assert(sequenceID == im['seq_id'])
        
        # If we have bounding boxes or an explicit empty label, create image-level annotations    
        if relPath in relPathToMetadataSet2:
            
            nFoundMetadata += 1
            
            # filename, number of bounding boxes, labeled boxes (five values per box)
            imageMetadata = relPathToMetadataSet2[relPath]
            
            # Make sure the relative filename matches, allowing for the fact that
            # some of the filenames in the metadata aren't quite right
            fn = os.path.normpath(imageMetadata[0])
            s = os.path.normpath(relPath.replace('Set2\\',''))
            assert(fn == s)
            
            nBoxes = int(imageMetadata[1])
            
            if nBoxes == 0:
                
                ann = {}
                ann['id'] = str(uuid.uuid1())
                ann['image_id'] = im['id']
                ann['category_id'] = emptyCat['id']
                ann['sequence_level_annotation'] = False
                annotationsSet2.append(ann)
                nImageLevelEmpties += 1
                
            else:
                
                for iBox in range(0,nBoxes):
                                    
                    boxCoords = imageMetadata[2+(iBox*5):7+(iBox*5)]
                    boxCoords = list(map(int, boxCoords))
                    
                    boxClass = boxCoords[0]
                    boxCat = categoryMappingsSet2[boxClass]
                    categoryId = boxCat['id']
                    
                    boxCoords = boxCoords[1:]
                    
                    # Bounding box values are in absolute coordinates, with the origin 
                    # at the upper-left of the image, as [xmin1 ymin1 xmax1 ymax1].
                    #
                    # Convert to floats and to x/y/w/h
                    bboxW = boxCoords[2] - boxCoords[0]
                    bboxH = boxCoords[3] - boxCoords[1]
                    
                    box = [boxCoords[0], boxCoords[1], bboxW, bboxH]
                    box = list(map(float, box))
                    
                    ann = {}
                    ann['id'] = str(uuid.uuid1())
                    ann['image_id'] = im['id']
                    ann['category_id'] = categoryId
                    ann['sequence_level_annotation'] = False
                    ann['bbox'] = box
                    annotationsSet2.append(ann)
                    nTotalBoxes += 1
                    
                # ...for each box
                
            # ...if we do/don't have boxes for this image
            
        # Else create a sequence-level annotation
        else:
            
            if sequenceID not in set2SequenceToClass:
                sequenceClass = 'empty'
            else:
                sequenceClass = set2SequenceToClass[sequenceID]
            category = sequenceClassCategoryMappingsSet2[sequenceClass]
            categoryId = category['id']
            
            ann = {}
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = im['id']
            ann['category_id'] = categoryId
            ann['sequence_level_annotation'] = True
            annotationsSet2.append(ann)
            nSequenceLevelAnnotations += 1
            
    # ...for each image
            
    print('Finished processing set 2, found metadata for {} of {} images, created {} annotations and {} boxes in {} categories'.format(
            nFoundMetadata,len(set2ImageIDs),len(annotationsSet2),nTotalBoxes,len(categoriesSet2)))
    
    assert len(annotationsSet2) == nSequenceLevelAnnotations + nTotalBoxes + nImageLevelEmpties
    assert len(set2ImageIDs) == nSequenceLevelAnnotations + nFoundMetadata
        
    
#%% The 'count' field isn't really meaningful, delete it

# It's really the count of image-level annotations, not total images assigned to a class
for d in categoriesSet1:
    del d['count']
    
    
#%% Write output .json files

data = {}
data['info'] = infoSet1
data['images'] = imagesSet1
data['annotations'] = annotationsSet1
data['categories'] = categoriesSet1
json.dump(data, open(outputJsonFilenameSet1,'w'), indent=4)    
print('Finished writing json to {}'.format(outputJsonFilenameSet1))

if process_set2:
    data = {}
    data['info'] = infoSet2
    data['images'] = imagesSet2
    data['annotations'] = annotationsSet2
    data['categories'] = categoriesSet2
    json.dump(data, open(outputJsonFilenameSet2,'w'), indent=4)
    print('Finished writing json to {}'.format(outputJsonFilenameSet2))


#%% Sanity-check final set 1 .json file

from data_management.databases import sanity_check_json_db
options = sanity_check_json_db.SanityCheckOptions()
sortedCategories,data,_ = sanity_check_json_db.sanity_check_json_db(outputJsonFilenameSet1, options)
sortedCategories

# python sanity_check_json_db.py --bCheckImageSizes --baseDir "E:\wildlife_data\missouri_camera_traps" "E:\wildlife_data\missouri_camera_traps\missouri_camera_traps_set1.json"


#%% Generate previews

from md_visualization import visualize_db

output_dir = os.path.join(baseDir,'preview')

options = visualize_db.DbVizOptions()
options.num_to_visualize = None
options.sort_by_filename = False
options.classes_to_exclude = None
options.trim_to_images_with_bboxes = True
options.parallelize_rendering = True

htmlOutputFile,_ = visualize_db.process_images(outputJsonFilenameSet1,output_dir,baseDir,options)
os.startfile(htmlOutputFile)

